## a wrapper script to summarize the results of ancIBD (combine results from all 22 autosomes)
from ancIBD.IO.ind_ibd import combine_all_chroms
from ancIBD.IO.ind_ibd import create_ind_ibd_df
import argparse
import os
from pathlib import Path

def main():
    parser = argparse.ArgumentParser(description='Run ancIBD.')
    parser.add_argument('--tsv', action="store", dest="tsv", type=str, required=True,
                        help="base path to the individual IBD files.")
    parser.add_argument('--ch', action="store", dest="ch", type=str, required=False, default='1-22', help='chromosome number, expressed in the format chrom-chrom, e.g, 1-22). The default is 1-22.')
    parser.add_argument('--bin', action="store", dest="bin", type=str, required=False, default='8,12,16,20', help='length bin over which IBD sharing summary statistics for pairs of samples will be calculated. Default is 8,12,16,20.')
    parser.add_argument('--snp_cm', action="store", dest="snp_cm", type=float, required=False, default=220, help='minimum number of SNPs per centimorgan for a segment to be considered. The default is 220 to reduce false positive rates.')
    parser.add_argument('--out', action="store", dest="out", type=str, required=False, help='output folder to store results. If not specified, the results will be stored in the current directory.')
    args = parser.parse_args()

    ch1, ch2 = args.ch.split('-')
    chs = range(int(ch1), int(ch2)+1)
    oDir = args.out
    if oDir is None:
        # set output directory to current directory
        oDir = os.getcwd()

    combine_all_chroms(chs=chs, folder_base=args.tsv,
                   path_save=os.path.join(oDir, 'ch_all.tsv'))
    
    #### now save a summary of IBD sharing between all pairs of individuals
    min_cms = [float(l) for l in args.bin.split(',')]
    create_ind_ibd_df(ibd_data = os.path.join(oDir, 'ch_all.tsv'),
                      min_cms = min_cms, snp_cm = args.snp_cm, min_cm = min_cms[0], sort_col = 0,
                      savepath = os.path.join(oDir, "ibd_ind.tsv"))


import matplotlib.pyplot as plt
from matplotlib import gridspec
import numpy as np
import h5py as h5py

from matplotlib import rcParams
rcParams['font.family'] = 'sans-serif'   # Set the default
rcParams['font.sans-serif'] = ['Arial']  # Make sure to have the font installed (it is on cluster for Harald)

def plot_posterior(ax=0, morgan=[], post=[], het=[], het_m=[], 
                   df_ibd=[], df_truth=[], state=0, figsize=(12,3), 
                   xlim=[], ylim=[-0.08,1.27], ylabel="Posterior", xlabel="Position",
                   c="maroon", het_c="gray", c_truth="green", ms=1,
                   lw=3, lw_ibd=10, c_ibd="slateblue", y_ibd=1.2, dpi=400, 
                   fs_l=12, show=True, min_cm=4, title="", savepath=""):
    """Plot Posterior [k,l] array. If morgan given, plot in centimorgan.
    Can then also plot hapROH formatted IBD blocks (df_ibd).
    And plot ground truth hapROH formatted IBD blocks (df_truth).
    If het is given [array boolean], plot het using het_m coordinates"""
    if ax==0:
        plt.figure(figsize=figsize)
        ax=plt.gca()
    if len(morgan)==0:
        morgan = np.arange(np.shape(post)[1])
    ax.plot(morgan*100, post[state,:], color=c, lw=lw)
    ax.set_yticks([0., 0.2, 0.4, 0.6, 0.8, 1.0])
    ### Do optional plotting
    # Hets
    if len(xlabel)>0:
        ax.set_xlabel(xlabel, fontsize=fs_l)
    if len(ylabel)>0:
        ax.set_ylabel(ylabel, fontsize=fs_l)
    if len(xlim)>0:
        ax.set_xlim(xlim)
    if len(ylim)>0:
        ax.set_ylim(ylim)
    if len(het)>0:
        plot_hets(ax, het_m, het, ms=ms, het_c=het_c, fs_l=fs_l, ylim=ylim)
    if len(title)>0:
        ax.set_title(title, fontsize=fs_l)
    if len(df_ibd)>0:
        df_ibd = df_ibd[df_ibd["lengthM"] > min_cm/100]  # Filter out long enough calls
        ax.hlines(y=[y_ibd]*len(df_ibd), xmin=100 * df_ibd["StartM"], xmax= 100 * df_ibd["EndM"], 
                        colors=c_ibd, linewidth=lw_ibd)
    if len(df_truth)>0:  # Plot Ground truth dataframe
        ### Plot them
        plt.hlines(y=[1.12]*len(df_truth), xmin=100 * df_truth["IBD_Begin"], 
                   xmax=100 * df_truth["IBD_End"], colors=c_truth, linewidth=lw_ibd)
        
    if len(savepath)>0:
        plt.savefig(savepath, bbox_inches ='tight', pad_inches = 0, dpi=dpi)
        print(f"Saved to {savepath}")
    if show:
        plt.show()
    else: 
        return ax
    
def plot_hets(ax, het_m, het, alpha=0.3, ms=1, het_c="slateblue",
              ylabel = "Opp. Homozygotes (no/yes)", fs_l=12, ylim=[]):
    """Plot Heterozygote Markers onto Axis"""
    ax2 = ax.twinx()
    ax2.plot(het_m*100, (het * 1.1 - 0.05), "o", ms=ms, alpha=alpha, zorder=0, color=het_c)
    ax2.set_yticks([-0.05, 1.05])
    ax2.set_yticklabels([])
    ax2.set_ylabel(ylabel, fontsize=fs_l, color=het_c)
    if len(ylim)>0:
        ax2.set_ylim(ylim)
    
########################################################
    
def plot_posterior_panel(post=[], figsize=(12,9), c="maroon", c_hw="g",
                         hspace=0.12, wspace=0.15, xlim=[], ylim=[-0.05,1.05], 
                         lw=3, fs_l=12, ch=0, savepath=""):
    """Plot Posterior [k,l] array.
    ch: If bigger 0, load 1240K map postion"""
    
    if ch>0:
        m = get_map(ch=ch)
        assert(len(m)==np.shape(post)[1]) # Sanity Check
    else:
        m=[]
        
    fig = plt.figure(figsize=figsize)
    gs = gridspec.GridSpec(ncols=2, nrows=3, figure=fig, 
                           wspace=wspace, hspace=hspace)
    ax = fig.add_subplot(gs[0, 0])

    pos = [[1,0], [1,1], [2,0], [2,1]]
    axs= [fig.add_subplot(gs[i[0], i[1]]) for i in pos]
    labels = ["0-0", "1-0", "0-1", "1-1"]
    labels = ["Posterior:" + l + " IBD" for l in labels]
    
    plot_posterior(ax, morgan=m, post=post, state=0, c=c_hw, 
                       xlim=xlim, ylim=ylim, xlabel="",
                       ylabel="Posterior: Non-IBD", lw=lw, show=False)
    
    for i, ax in enumerate(axs):
        plot_posterior(ax, morgan=m, post=post, state=i+1, c=c, 
                       xlim=xlim, ylim=ylim, xlabel="",
                       ylabel=labels[i], lw=lw, show=False)
    if len(savepath)>0:
        plt.savefig(savepath, bbox_inches ='tight', pad_inches = 0, dpi=400)
        print(f"Saved to {savepath}")
    plt.show()
    
def get_map(path_h5= "./data/hdf5/1240k_v43/ch", ch=3, 
            col_map="variants/MAP"):
    """Get Map position in Morgan"""
    path_load = f"{path_h5}{ch}.h5"
    f = h5py.File(path_load, "r") # Load for Sanity Check. See below!
    m = f[col_map][:]
    return m


"""
Class containing functions to load data needed for HMM IBD Run.
Returns all relevant data for HMM in standardized format:
1) Haplotype Probabilities (for ancestral allele)
2) Allele Frequencies 
3) Genetic Map
@ Author: Harald Ringbauer, September 2020
"""

import pickle as pickle
import numpy as np
import os as os
import h5py as h5py

###############################
###############################

class LoadData(object):
    """Class to load data in uniform format"""
    path = "" # Path to load from (usually folder)
    output = True # Whether to print output
    
    def __init__(self, path=""):
        """Can remember path if given"""
        self.path=path
    
    def return_p(self, **kwargs):
        """Return array of Allele Frequencies [l]"""
        raise NotImplementedError("Implement This in specific subclass.")

    def return_map(self, **kwargs):
        """Return genetic map [l] in Morgan"""
        raise NotImplementedError("Implement This in specific subclass.")
        
    def return_haplotypes_ll(self, **kwargs):
        """Return haplotype likelihoods [4,l,2]"""
        raise NotImplementedError("Implement This in specific subclass.")
        
    def load_all_data(self, **kwargs):
        """Load all haplotype likelihoods
        haplotype likelihoods [2*n,l,2]
        derived allele frequencies [l]
        map in Morgan [l]"""
        htsl = self.return_haplotypes_ll()
        p = self.return_p()
        m = self.return_map()
        self.check_valid_data(htsl, p, m)
        return htsl, p, m
    
    def check_valid_data(self, htsl, p, m):
        """Check whether data in valid format"""
        assert(np.shape(htsl)[1]==len(p))
        assert(np.sum(np.isnan(htsl))==0) # Check that no NAN exists
        assert(len(p)==len(m))
        pass 
        
    def set_params(self, **kwargs):
        """Set the Parameters.
        Takes keyworded arguments"""
        for key, value in kwargs.items():
            setattr(self, key, value)
            
###############################
### Class for Simulated date

class LoadSimulated(LoadData):
    """Class to load simulated data saved in standard format."""
    r_gap = 1.0
    r_path = ""
    
    def load_all_data(self, **kwargs):
        ### Calculate Paths
        haplo_path = os.path.join(self.path, "haplo_ll.tsv")
        p_path = os.path.join(self.path, "p.tsv")
        
        ### Load the Data
        #htsl = pickle.load(open(haplo_path, "rb")) 
        htsl = np.loadtxt(haplo_path, delimiter="\t", dtype="float")
        p = np.loadtxt(p_path, delimiter="\t", dtype="float")
        if len(self.r_path)==0:
            #m = np.ones(len(p), dtype="float") * self.r_gap
            m = np.arange(len(p), dtype="float") * self.r_gap # Load absolute Positions
        
        self.check_valid_data(htsl, p, m)
        return htsl, p, m
    
###############################
### Class for HDF5 data

class LoadHDF5(LoadData):
    """Class to load HDF data for 2 individuals in standard format."""
    path_h5=""
    iids = []
    ch = 3   # Which chromosome to load
    min_error=1e-5 # Minimum Probability of genotyping error
    p_col = "variants/AF_ALL" # The hdf5 column with der. all freqs
    
    def return_map(self, f):
        """Return the recombination map"""
        m = f["variants/MAP"][:]
        return m
    
    def return_p(self, f):
        """Return array of Allele Frequencies [l]
        self.p_col: The HDF5 field with the allele frequencies
        If not given, use default p=0.5
        """
        if len(self.p_col)==0:
            p = 0.5 * np.ones(len(f["variants/MAP"]))
        else:
            p = f[self.p_col][:] # Load the Allele Freqs from HDF5
            if len(p.shape) > 1:
                p = p[:,0]
        return p
        
    def get_individual_idx(self, f, iid="", f_col="samples"):
        """Return index of individual iid"""
        samples = f[f_col].asstr()[:]
        idx = (samples == iid)
        if np.sum(idx)!=1:
            raise RuntimeWarning(f"{np.sum(idx)} entries found for {iid}")
        assert(np.sum(idx)>0) # Sanity Check
        idx=np.where(idx)[0][0]
        return idx  
    
    def get_haplo_prob(self, f, idx):
        """Get haploid ancestral probability for indivual [2,l]"""
        h1 = f["calldata/GT"][:,idx,:].T
        m = np.max(f["calldata/GP"][:,idx,:], axis=1)
        m = np.minimum(m,  1 - self.min_error)
        h1 = (1 - h1) * m + h1 * (1 - m) # Probability of being ancestral
        return h1
        
    def load_all_data(self, **kwargs):
        """ Return haplotype likelihoods [4,l] for anc. allele
        derived allele frequencies [l]
        map in Morgan [l]"""
        path_h5_ch = f"{self.path}{self.ch}.h5"
        with h5py.File(path_h5_ch, "r") as f:
            m = self.return_map(f)
            p = self.return_p(f)
            
            idx1 = self.get_individual_idx(f, self.iids[0])
            idx2 = self.get_individual_idx(f, self.iids[1])
            h1 = self.get_haplo_prob(f, idx1)
            h2 = self.get_haplo_prob(f, idx2)
            htsl = np.concatenate((h1,h2), axis=0)
        
        self.check_valid_data(htsl, p, m)
        return htsl, p, m, self.iids
    
class LoadHDF5Multi(LoadHDF5):
    """Class to load HDF5 data for multiple individuals.
    Default now also for only pairs of individuals."""
    path_h5=""
    iids = []
    ch = 3   # Which chromosome to load
    min_error = 1e-5 # Minimum Probability of genotyping error  
    p_col = "" # The hdf5 column with der. all freqs. If given, use the field
    # If "default" use p=0.5 everywhere
    # default value for p_col in my hdf5s: variants/AF_ALL
    
    def get_haplo_prob(self, f, idcs):
        """Get haploid ancestral probability for n individuals 
        Return [n,l,2] array"""
        h1 = f["calldata/GT"][:,idcs,:] ### Get l, n, 2
        l,n,_ = np.shape(h1) # Get the nr loci
        m = np.max(f["calldata/GP"][:,idcs,:], axis=2)  
        m = np.minimum(m,  1 - self.min_error) # Max Cap of certainty
        h1 = (1-h1) * m[:,:,None] + h1 * (1 - m[:,:,None]) # Probability of being ancestral
        h1 = np.swapaxes(h1, 0, 1) # ->n,l,2
        h1 = np.swapaxes(h1, 1, 2) #-> n,2,l
        h1 = h1.reshape((2*n,l))
        #h1 = np.swapaxes(h1, 0, 1)
        return h1
    
    def filter_valid_data(self, hts, p, m):
        """Filter to SNPs with fully valid data. 
        Return filtered data."""
        idx = ~(np.isnan(hts).any(axis=0)) # Flag all markers that are not null
        
        ### Output and Raise warning if too much data missing:
        if self.output:
            print(f"Filtering to {np.sum(idx)}/{len(idx)} SNPs with GP data (on target iids)")   
        if np.mean(idx)<1:  # Missing Data detected
            print(f"Attention: Some data in GP field is missing. Ideally, all GP entries are set.")   
        if np.mean(idx)<0.8:  # Less than 80 percent of data there
            raise RuntimeWarning(f"Too much data missing: {np.sum(idx)}/{len(idx)} SNPs are NULL for GP")
            
        return hts[:,idx], p[idx], m[idx]
    
    def load_all_data(self, **kwargs):
        """ Return haplotype likelihoods [n*2,l] for anc. allele.
        along first axis: 2*i, 2*(i+1) haplotype of ind i
        derived allele frequencies [l]
        map in Morgan [l]"""
        path_h5_ch = f"{self.path}{self.ch}.h5"
        with h5py.File(path_h5_ch, "r") as f:
            m = self.return_map(f)            
            idcs = np.array([self.get_individual_idx(f, iid) for iid in self.iids])
            sort = np.argsort(idcs)   # Get the sorting Indices [has to go low to high]
            samples = self.iids[sort] # Get them in sorted order
            hts = self.get_haplo_prob(f, idcs[sort])
        
            if len(self.p_col)>0:
                p = self.get_p_hdf5(f, self.p_col)
            else:
                p = self.get_p(hts)  # Calculate Mean allele frequency from sample subset
                
        ### Filter to Valid data
        hts, p, m = self.filter_valid_data(hts, p, m)
        
        self.check_valid_data(hts, p, m)
        return hts, p, m, samples
    
    def get_p(self, htsl):
        """Get Allele frequency from haplotype probabilities.
        Return array of derived allele freqs [l]"""
        p_anc = np.mean(htsl, axis=0) # Take the expected AF of ancestral.
        p_der = 1 - p_anc             # Get the derived allele frequency
        return p_der
    
    def get_p_hdf5(self, f, col):
        """Get allele frequs from HDF f 
        in dataset col.
        Return array of derived allele freqs [l]"""
        if col=="default":
            p = 0.5 * np.ones(len(f["variants/MAP"]))
        else:
            p = f[col][:] # Load the Allele Freqs from HDF5
            
        ### Filter if p is matrix (e.g. sometimes for three alt alleles)
        if len(np.shape(p))==2:
            p=p[:,0]
        elif len(np.shape(p))!=1:
            raise RuntimeWarning(f"Allele Frequency field in h5 {col} has invalid dimensions!")
            
        return p

class LoadH5Multi2(LoadHDF5Multi):
    """Update to more accurate genotype probabilities
    from diploid to haploid."""
    path_h5=""
    iids = []
    ch = 3   # Which chromosome to load
    min_error = 1e-3  # Minimum Probability of genotyping error  
    pph_error = 1e-2 #1e-3 # Point Phase Error
    p_col = "" # The hdf5 column with der. all freqs. If given, use the field
    # If "default" use p=0.5 everywhere
    # default value for p_col in my hdf5s: variants/AF_ALL
    
    def get_haplo_prob(self, f, idcs):
        """Get haploid ancestral probability for n individuals 
        Return [n,l,2] array. Calculated from GP and GT
        in the proper way."""
        
        gt = f["calldata/GT"][:,idcs,:] ### Get l, n, 2 array of gt
        l,n,_ = np.shape(gt) # Get the nr loci
        gp = f["calldata/GP"][:,idcs,:] ### Get l, n, 3 array of gp
        
        ### The homozygotes
        g00 = gp[:,:,0]  # homo 00 prob.
        ### Heterozygote prob
        gp1 = gp[:,:,1]
        ## g11 not needed
        
        ### The max GT probabilities
        g01 = gp1[:,:] * (gt[:,:,0]==0) * (gt[:,:,1]==1) # het 01 prob.
        g10 = gp1[:,:] * (gt[:,:,0]==1) * (gt[:,:,1]==0) # het 10 prob.
        
        ### The default probability from the non-max GT ones
        # Assume it is split up 50/50 between 01 10 states
        idx = (gt[:,:,0]==0) * (gt[:,:,1]==0)
        g01[idx] = gp1[idx]/2
        g10[idx] = gp1[idx]/2
        
        idx = (gt[:,:,0]==1) * (gt[:,:,1]==1)
        g01[idx] = gp1[idx]/2
        g10[idx] = gp1[idx]/2
        ### The added probability from the non-max GT ones
        
        # Add point phasing error  here (swap g01 and g10 with prob. x)
        g01t = g01 * (1 - self.pph_error) + g10 * self.pph_error
        g10t = g10 * (1 - self.pph_error) + g01 * self.pph_error
        
        h1 = np.zeros((l,n,2), dtype=np.float64)
        h1[:,:,0] = g00 + g01t 
        h1[:,:,1] = g00 + g10t
        
        h1 = np.swapaxes(h1, 0, 1) #  l,n,2->n,l,2
        h1 = np.swapaxes(h1, 1, 2) # -> n,2,l
        h1 = h1.reshape((2*n,l))
        
        ### Add minimum error to avoid overly extreme confidence genotypes
        h1 = np.maximum(h1, self.min_error)
        h1 = np.minimum(h1, 1-self.min_error)
        
        if self.output:
            print(f"Min. Error added: {self.min_error}")
            print(f"Phase. Error added: {self.pph_error}")
        
        return h1
    
###############################    
###############################
### Factory Method
    
def load_loaddata(l_model="simulated", path="", **kwargs):
    """Factory method to return the right loading Model"""
    if l_model == "simulated":
        l_obj = LoadSimulated(path=path, **kwargs)
    elif l_model == "hdf5double":  # use overall allele frequency
        l_obj = LoadHDF5(path=path, **kwargs)
    elif l_model == "hdf5":
        l_obj = LoadHDF5Multi(path=path, **kwargs)
    elif l_model == "h5":  # Latest model, as described in paper.
        l_obj = LoadH5Multi2(path=path, **kwargs)
    else:
        raise NotImplementedError("Loading Model not found!")
    return l_obj


"""
Function for running hapBLOCK on single chromsome, with all relevant keywords.
Function to run hapBLOCK on a full individual and full reference Dataset,
with all relevant keywords.
@Author: Harald Ringbauer, 2020
"""

import numpy as np
import multiprocessing as mp
import pandas as pd
import itertools as it
from time import time
import sys as sys
import os as os

#sys.path.append("/n/groups/reich/hringbauer/git/hapBLOCK/python3/") 
from ancIBD.main import HMM_Full  # To run the main plotting.
from ancIBD.plot.plot_posterior import plot_posterior # to plot the posterior.
from ancIBD.IO.h5_load import get_opp_homos_f

def hapBLOCK_chrom(folder_in="./data/hdf5/1240k_v43/ch", iids = ["", ""], 
                   ch=2, folder_out="", output=False, prefix_out="", logfile=False,
                   l_model="h5", e_model="haploid_gl", h_model="FiveStateScaled", 
                   t_model="standard", p_col="variants/AF_ALL", 
                   ibd_in=1, ibd_out=10, ibd_jump=400, ibd_jump2=0.5, min_cm=2, min_error=1e-3,
                   cutoff_post=0.99, max_gap=0.0075):
    """Run IBD for pair of Individuals.
    folder_in: hdf5 path up to chromosome.
    iids: List of IIDs to compare [length 2]
    folder_out: Where to save the hapBLOCK output to
    min_cm: Minimal block length to call and save [cM]
    savepath: Where to save the IBD plot to.
    p_col: The dataset to use in hdf5 for der. AF. If default use p=0.5.
    min_error: Caps min/max prob. haplotype being derived to min_error/1-min_error when loading data
    If empyt use in sample AF.
    Return df_ibd, posterior, map, tot_ll"""
    iids = np.array(iids) # For better props when indexing
    assert(len(iids)==2) # Sanity Check of Input IIDs
    h = HMM_Full(folder_in=folder_in, l_model=l_model, t_model=t_model, 
                     e_model=e_model, h_model = h_model,
                     output=output, load=True)

    if t_model == 'asymmetric':
        h.t_obj.set_params(ibd_in = ibd_in, ibd_out = ibd_out, ibd_jump = ibd_jump, ibd_jump2 = ibd_jump2)
    else:
        h.t_obj.set_params(ibd_in = ibd_in, ibd_out = ibd_out, ibd_jump = ibd_jump)
        
    h.l_obj.set_params(iids=iids, ch=ch, p_col=p_col, min_error=min_error)
    h.p_obj.set_params(ch=ch, min_cm=min_cm, cutoff_post=cutoff_post, max_gap=max_gap)
    #post, r_vec, fwd, bwd, tot_ll = h.run_fwd_bwd()
    post, r_vec =  h.run_fwd_bwd(full=False)
    df_ibd, _, _ = h.p_obj.call_roh(r_vec, post)
    
    if len(folder_out)>0:
        folder_out = h.prepare_path(folder_out, ch=ch, prefix_out=prefix_out, logfile=logfile)
        save_path = os.path.join(folder_out, f"ch{ch}.tsv")
        h.p_obj.save_ibd_df(df_ibd=df_ibd, save_path = save_path)
    return df_ibd, post, r_vec


def prep_param_list_chrom(folder_in, iids = [], ch=3,
                    folder_out="", output=True, logfile=False, prefix_out="default/",
                    l_model="h5", e_model="haploid_gl", h_model="FiveStateScaled", 
                    t_model="standard", p_col="variants/AF_ALL", ibd_in=1, ibd_out=1, ibd_jump=500, min_cm=2,
                    cutoff_post=0.99, max_gap=0.0):
    """Prepare parameter lists for multirun of hapBLOCK_chrom. Ideal for multi-processing,
    as it gives a list of parameters - one for each iid pair."""
    params = [[folder_in, iid2, ch, folder_out, output, prefix_out, logfile, l_model, e_model,
              h_model, t_model, p_col, ibd_in, ibd_out, ibd_jump, min_cm, cutoff_post, max_gap] for iid2 in iids]
    assert(len(params[0])==18)
    return params

#################################################################################
#################################################################################

def hapBLOCK_chroms(folder_in="./data/hdf5/1240k_v43/ch", iids = [], run_iids=[],
                   ch=2, folder_out="", output=False, prefix_out="", logfile=False,
                   l_model="h5", e_model="haploid_gl", h_model="FiveStateScaled", 
                   t_model="standard", p_model="hapROH", p_col="variants/AF_ALL", 
                   ibd_in=1, ibd_out=10, ibd_jump=400, ibd_jump2=0.5, min_cm=2,
                   cutoff_post=0.99, max_gap=0.0075, processes=1):
    """Run IBD for list of Individuals, and saves their IBD csv into a single 
    output folder.
    folder_in: hdf5 path up to chromosome.
    iids: List of IIDs to load [k indivdiuals]
    run_iids: If given: list of IID pairs to run. If not run all pairs
    folder_out: Where to save the hapBLOCK output to
    min_cm: Minimal block length to call and save [cM]
    savepath: Where to save the IBD plot to.
    Return df_ibd, posterior, map, tot_ll"""
    ### Run all pairs if empty
    iids = np.array(iids) # For better props when indexing
    if not len(set(iids))==len(iids): # Check whether duplicates
        raise RuntimeWarning("Duplicate IIDs detected!")
    if len(run_iids)==0:
        run_iids = it.combinations(iids, 2)
        
    ### Load all the objects
    h = HMM_Full(folder_in=folder_in, l_model=l_model, t_model=t_model, 
                     e_model=e_model, h_model = h_model, p_model=p_model,
                     output=output, load=True)
    if t_model == 'asymmetric':
        h.t_obj.set_params(ibd_in = ibd_in, ibd_out = ibd_out, ibd_jump = ibd_jump, ibd_jump2 = ibd_jump2)
    else:
        h.t_obj.set_params(ibd_in = ibd_in, ibd_out = ibd_out, ibd_jump = ibd_jump)
    h.l_obj.set_params(iids=iids, ch=ch, p_col=p_col)
    h.p_obj.set_params(ch=ch, min_cm=min_cm, cutoff_post=cutoff_post, max_gap=max_gap)
    
    ### Load all data
    htsl, p, r_vec, samples =  h.l_obj.load_all_data()
    
    ### Load transition matrix
    t_mat = h.t_obj.full_transition_matrix(r_vec, n=4, submat33 = h.submat33)
    
    ### loop over all Run Pair Individuals
    df_ibds = []
    for iid1,iid2 in run_iids:
        i1 = get_sample_index(samples, iid1)
        i2 = get_sample_index(samples, iid2) 
        idcs = [i1*2, i1*2+1, i2*2, i2*2+1] # Get the right indices
        e_mat =  h.e_obj.give_emission_matrix(htsl[idcs,:], p)
        post =  h.fwd_bwd(e_mat, t_mat, in_val =  h.in_val, 
                            full=False, output= h.output)
        df_ibd, _, _ = h.p_obj.call_roh(r_vec, post, iid1, iid2)
        df_ibds.append(df_ibd)
    
    df_ibds = pd.concat(df_ibds)
    
    if len(folder_out)>0:
        folder_out = h.prepare_path(folder_out, ch=ch, prefix_out=prefix_out, logfile=logfile)
        save_path = os.path.join(folder_out, f"ch{ch}.tsv")
        h.p_obj.save_ibd_df(df_ibd=df_ibds, save_path = save_path)

    return df_ibds

def get_sample_index(iids, sample):
    """Get Index of sample - check if really there"""
    idx = np.where(iids[:]==sample)[0]
    assert(len(idx)==1)
    return idx[0]

############################################################################
### For timing

def hapBLOCK_times(folder_in="./data/hdf5/1240k_v43/ch", iids = [], run_iids=[],
                   ch=2, folder_out="", output=False, prefix_out="", logfile=False,
                   l_model="h5", e_model="haploid_gl", h_model="FiveStateScaled", 
                   t_model="standard", p_col="variants/AF_ALL", ibd_in=1, ibd_out=10, ibd_jump=400, min_cm=2,
                   cutoff_post=0.99, max_gap=0.0075, processes=1):
    """Run IBD for list of Individuals, and returns runtimes. Same as hapBLOCK_chroms (see docstring there) but also return runtimes. USED ONLY FOR BENCHMARKING
    Return df_ibd, 
    t1 [start time], t2 [after loading], t3 [after calculating trans rates], t4 [before saving]
    """
    ### Run all pairs if empty
    iids = np.array(iids) # For better props when indexing
    if not len(set(iids))==len(iids): # Check whether duplicates
        raise RuntimeWarning("Duplicate IIDs detected!")
    if len(run_iids)==0:
        run_iids = it.combinations(iids, 2)
        
    t1 = time()
    
    ### Load all the objects
    h = HMM_Full(folder_in=folder_in, l_model=l_model, t_model=t_model, 
                     e_model=e_model, h_model = h_model,
                     output=output, load=True)
    h.t_obj.set_params(ibd_in = ibd_in, ibd_out = ibd_out, ibd_jump = ibd_jump)
    h.l_obj.set_params(iids=iids, ch=ch, p_col=p_col)
    h.p_obj.set_params(ch=ch, min_cm=min_cm, cutoff_post=cutoff_post, max_gap=max_gap)
    
    ### Load all data
    htsl, p, r_vec, samples =  h.l_obj.load_all_data()
    
    t2 = time()
    
    ### Load transition matrix
    t_mat = h.t_obj.full_transition_matrix(r_vec, n=4, submat33 = h.submat33)
    
    t3 = time()
    
    
    ### loop over all Run Pair Individuals
    df_ibds = []
    for iid1,iid2 in run_iids:
        i1 = get_sample_index(samples, iid1)
        i2 = get_sample_index(samples, iid2) 
        idcs = [i1*2, i1*2+1, i2*2, i2*2+1] # Get the right indices
        e_mat =  h.e_obj.give_emission_matrix(htsl[idcs,:], p)
        post =  h.fwd_bwd(e_mat, t_mat, in_val =  h.in_val, 
                            full=False, output= h.output)

        df_ibd, _, _ = h.p_obj.call_roh(r_vec, post, iid1, iid2)
        df_ibds.append(df_ibd)
    
    df_ibds = pd.concat(df_ibds)
    
    t4 = time()
    
    if len(folder_out)>0:
        folder_out = h.prepare_path(folder_out, ch=ch, prefix_out=prefix_out, logfile=logfile)
        save_path = os.path.join(folder_out, f"ch{ch}.tsv")
        h.p_obj.save_ibd_df(df_ibd=df_ibds, save_path = save_path)

    return df_ibds, t1, t2, t3, t4

############################################################################
### Run and plot in one go - for one pair of iids.

def run_plot_pair(path_h5="/n/groups/reich/hringbauer/git/hapBLOCK/data/hdf5/1240k_v43/ch", 
                  iids = ["", ""], ch=2, xlim=[], folder_out="", 
                  plot=False, path_fig="", output=False, exact=True,
                  ibd_in=1, ibd_out=10, ibd_jump=400, min_cm=2, 
                  cutoff_post=0.99, max_gap=0.0075, min_error=1e-3,
                  l_model="hdf5", e_model="haploid_gl", h_model="FiveStateScaled",
                  p_col="variants/AF_ALL",
                  title="", c="gray", c_hw="maroon", 
                  state=0, return_post=False, **kwargs):
    """Run and plot IBD for pair of Individuals.
    folder_out: Where to save the hapBLOCK output to
    iids: list of two iids [List of Length 2]
    path_fig: Where to save the IBD plot to [String]
    p_col: The dataset to use in hdf5 for der. AF. If default use p=0.5.
           If empty string use in sample AF.
    return_post: Whether to return posterior [Boolean]
    min_error: Caps min/max prob. haplotype being derived to min_error/1-min_error when loading data
    kwargs: Optional Keyword Arguments for Plotting (e.g. c_ibd)
    """
    assert(len(iids)==2) # Sanity Check of Input IIDs - as here it should be pairs
    
    df_ibd, post, r_vec = hapBLOCK_chrom(
           folder_in=path_h5, iids = iids,  ch=ch, folder_out=folder_out, 
           output=output, prefix_out="", logfile=False,
           l_model=l_model, e_model=e_model, h_model=h_model, 
           t_model="standard", p_col=p_col, ibd_in=ibd_in, ibd_out=ibd_out, ibd_jump=ibd_jump, 
           min_error=min_error, min_cm=min_cm, cutoff_post=cutoff_post, max_gap=max_gap)
        
    if plot:
        if len(title)==0:
            title = f"ancIBD v0.5, {iids[0]} - {iids[1]}, Chr. {ch}"
            
        ### Load the data from the HDF5
        o_homos, m = get_opp_homos_f(iid1=iids[0], iid2=iids[1], 
                                     f_path=path_h5, ch=ch, exact=exact)
        print(f"Plotting {len(r_vec)} markers")
        plot_posterior(post=post, morgan=r_vec, df_ibd=df_ibd, 
                       het=o_homos, het_m=m, state=state,
                       min_cm=min_cm, title=title, xlim=xlim, show=True, 
                       savepath=path_fig, xlabel="Chromosome Position [cM]", **kwargs)
    if return_post:
        return post


"""
Class containing functions to calculate Emission Probabilities.
Contains Sub-Classes for various types of data, 
as well as a factory Method to return the 
correct subclass
@ Author: Harald Ringbauer, September 2020
"""

import numpy as np
import itertools as it

###############################
###############################

class Emissions(object):
    """Class for emission probabilities
    Has methods to return emission probabilities"""

    def give_emission_matrix(self, **kwargs):
        """Return Emission Matrix - for every possible set of states"""
        raise NotImplementedError("Implement This in specific subclass.")
    
    def give_emission_log(self, **kwargs):
        """Return the full emission Probability directly in Log Space. 
        By default just log the emissiom matrix. Can be overwritten for
        nicer computational properties"""
        m = self.give_emission_matrix(**kwargs)
        #assert(np.min(m)>0)
        return np.log(m)

    def set_params(self, **kwargs):
        """Set the Parameters.
        Takes keyworded arguments"""
        for key, value in kwargs.items():
            setattr(self, key, value)
            
class HaplotypeSharingEmissions(Emissions):
    """Class for emission probabilities of diploid haplotypes who
    possibly share one haplotype"""
    
    def hw_prob_haplo_pp(self, ht_p, p):
        """Calculate HW Probabilitiy of haplotype gt
        ht_p: [l] Array of haplotype prob for ancestral, l locis,
        p: [l] Array of (derived) allele frequencies
        returns [l] vector of HW prob of gt"""
        prob = (1 - ht_p[:]) * p +  ht_p[:] * (1-p)
        return prob

    def hw_prob_haplo_share_pp(self, ht_p1, ht_p2, p):
        """Calculate probability of sharing haplotypes
        ht_p1, ht_p2: Array of haplotype likelihood for ancestral, l locis
        p: [l] Array of (derived) allele frequencies
        returns [l] vector of prob that shared haplotype"""
        s1 = ht_p1 * ht_p2 * (1-p) # Prob both ancestral
        s2 = (1-ht_p1) * (1 -ht_p2) * p # Prob both derived
        prob = s1 + s2
        return prob

    def hw_prob_haplos_pp(self, hts_p, p):
        """Calculate HW Probabilitiy of haplotype gt
        gt: [k,l] Array of haplotype prob for ancestral
        p: [l] Array of (derived) allele frequencies
        returns [l] vector of HW prob of gt"""
        prob = (1-hts_p[:,:]) * p +  hts_p[:,:] * (1-p)
        prob_tot = np.prod(prob, axis=0)
        return prob_tot

    def hw_probs_shared(self, hts_p, p, shared=(0,2), dtype="float"):
        """Give emission probabilities for shared state.
        hts_p: [4,l] Array of four ancestral haplotype probabilities
        p: [l] array of derived genotype probability.
        shared: tuple of length 2 giving the indices of the shared haplotypes"""
        not_shared = [i for i in range(0,4) if i not in shared]
        assert(len(not_shared)==2 & len(shared)==2)
        p_hw1 = self.hw_prob_haplo_pp(hts_p[not_shared[0],:], p=p)
        p_hw2 = self.hw_prob_haplo_pp(hts_p[not_shared[1],:], p=p)
        p_shared = self.hw_prob_haplo_share_pp(hts_p[shared[0],:],hts_p[shared[1],:], p=p)
        p = p_hw1*p_hw2*p_shared
        return p

    def give_emission_matrix(self, hts_p, p, dtype="float"):
        """Give emission Matrix for 5-state HMM.
        0th state: HW 1st-4th State: Haplotype Copying
        Input: p: [l] Array of (derived) allele frequencies
        hts_p: [4,l] Array of four ancestral haplotype probabilities.
        Return: emission matrix [5,l]."""
        l = np.shape(hts_p)[1]
        e_mat = np.zeros((5,l), dtype=dtype)
        e_mat[0,:] = self.hw_prob_haplos_pp(hts_p,p=p)
        for i, (j,k) in enumerate(it.product([0,1],repeat=2)):
            e_mat[i+1,:] = self.hw_probs_shared(hts_p, p=p, shared=[j,k+2])
        return e_mat
    
class HaplotypeSharingEmissions2(Emissions):
    """Class for emission probabilities of diploid haplotypes who
    possibly share one haplotype. Update to new emissions based
    on that output of Glimpse are posterior probabilities"""
    
    
    def hw_prob_haplo_share_pp(self, ht_p1, ht_p2, p, min_p = 0.001):
        """Calculate probability of sharing haplotypes
        ht_p1, ht_p2: Array of haplotype likelihood for ancestral, l locis
        p: [l] Array of (derived) allele frequencies
        returns [l] vector of prob that shared haplotype"""
        ### Cut off to minimal allele frequencies
        pt = np.maximum(p, min_p)
        pt = np.minimum(pt, 1-min_p)
        
        s1 = ht_p1 * ht_p2 / (1-pt) # Prob both ancestral
        s2 = (1-ht_p1) * (1-ht_p2) / pt # Prob both derived
        prob = s1 + s2
        return prob

    def hw_probs_shared(self, hts_p, p, shared=(0,2), dtype="float"):
        """Give emission probabilities for shared state.
        hts_p: [4,l] Array of four ancestral haplotype probabilities
        p: [l] array of derived genotype probability.
        shared: tuple of length 2 giving the indices of the shared haplotypes"""
        ### Sanity Check
        not_shared = [i for i in range(0,4) if i not in shared]
        assert(len(not_shared)==2 & len(shared)==2) # Sanity Check
        
        # Actual Calculation
        p_shared = self.hw_prob_haplo_share_pp(hts_p[shared[0],:],hts_p[shared[1],:], p=p)
        return p_shared

    def give_emission_matrix(self, hts_p, p, dtype="float"):
        """Give emission Matrix for 5-state HMM.
        0th state: HW 1st-4th State: Haplotype Copying
        Input: p: [l] Array of (derived) allele frequencies
        hts_p: [4,l] Array of four ancestral haplotype probabilities.
        Return: emission matrix [5,l]."""
        l = np.shape(hts_p)[1]
        e_mat = np.zeros((5,l), dtype=dtype)
        e_mat[0,:] = 1 # Because simply proportional to P(D)
        for i, (j,k) in enumerate(it.product([0,1],repeat=2)):
            e_mat[i+1,:] = self.hw_probs_shared(hts_p, p=p, shared=[j,k+2])
        return e_mat

class HaplotypeSharingEmissions3(HaplotypeSharingEmissions2):
    """
    Basically the same as HaplotypeSharingEmissions2, but with two additional states to account for IBD2 region
    """


    def give_emission_matrix(self, hts_p, p, dtype='float'):
        """
        Give emission matrix for 7-state HMM.
        0th state: non-IBD state
        1st-4th state: IBD1 state
        5-7th state: IBD2 state
        """
        l = np.shape(hts_p)[1]
        e_mat = np.zeros((7,l), dtype=dtype)
        e_mat[:5,:] = super().give_emission_matrix(hts_p, p, dtype)

        x1A = 1 - hts_p[0,:]
        x1B = 1 - hts_p[1,:]
        x2A = 1 - hts_p[2,:]
        x2B = 1 - hts_p[3,:]
        # IBD2 situation 1: 1A=2A, 1B=2B
        e_mat[5,:] = (1-x1A)*(1-x1B)*(1-x2A)*(1-x2B)/((1-p)**2) + \
                (1-x1A)*x1B*(1-x2A)*x2B/(p*(1-p)) + \
                x1A*(1-x1B)*x2A*(1-x2B)/(p*(1-p)) + \
                x1A*x1B*x2A*x2B/(p**2)
        
        # IBD2 situation 2: 1A=2B, 1B=2A
        e_mat[6,:] = (1-x1A)*(1-x1B)*(1-x2A)*(1-x2B)/((1-p)**2) + \
                (1-x1A)*x1B*x2A*(1-x2B)/(p*(1-p)) + \
                x1A*(1-x1B)*(1-x2A)*x2B/(p*(1-p)) + \
                x1A*x1B*x2A*x2B/(p**2)

        return e_mat




def load_emission_model(e_model="haploid_gl"):
    """Factory method to return the right Emission Model"""
    if e_model == "haploid_gl":
        e_obj = HaplotypeSharingEmissions()
    elif e_model == "haploid_gl2":
        e_obj = HaplotypeSharingEmissions2()
    elif e_model == 'IBD2':
        e_obj = HaplotypeSharingEmissions3()
    else:
        raise NotImplementedError("Emission Model not found!")
    return e_obj


"""
Class for calculating Transition Probabilities, i.e. 
infitesimal transition Matrices.
Contains Sub-Classes, as well as factory Method.
@ Author: Harald Ringbauer, 2021, All rights reserved
"""
import numpy as np


class Transitions(object):
    """Class for transition probabilities.
    Has methods to return them"""
    trans_mat = []    # The full transition Matrix
    r_map = []        # Legacy Variable
    output=True

    def __init__(self):
        """Initialize Class"""
        #self.calc_transitions()
        pass

    def calc_transition_rate(self, n=0):
        """Calulate and return Transition Matrix"""
        raise NotImplementedError("Implement This!")
        
    def give_transition_rate(self):
        """Give the transition_matrix"""
        return self.trans_mat

    def set_params(self, **kwargs):
        """Set the Values."""
        for key, value in kwargs.items():
            setattr(self, key, value)


class FiveStateTransitions(Transitions):
    """Implements the Model Transitions [units generally in Morgan]"""
    ibd_in = 0.0005     # The rate of jumping into IBD copying state
    ibd_out = 0.001     # The rate of jumping out of IBD state
    ibd_jump = 0.05     # The rate of jumping within IBD to other haplotype pair
    
    min_gap=1e-10 # Minimum Map Gap between two loci [Morgan]
    max_gap=0.05  # Maximum Map Gap between two loci [Morgan]

    def calc_transition_rate(self, submat33=True, n=4):
        """Return Transition Rate Matrix [k,k] to exponate.
        n: Number of symetric IBD states. Usually four (2x2 copying possibilities)
        submat33: Whether to only fill in the first 3 states """

        if submat33 == True:
            # Initialize Transition Matrix Only do 3 States (bc symmetry)
            t_mat = -np.ones((3, 3))
        else:
            t_mat = -np.ones((n + 1, n + 1))  # Initialize Transition Matrix

        t_mat[1:, 0] = self.ibd_out  # The rate of jumping out roh
        t_mat[0, 1:] = self.ibd_in / n  # Jumping into any ROH State
        t_mat[1:, 1:] = self.ibd_jump / n  # Jumping between ROH State

        # Do the Diagonal (do the usual model - for inf. substract 1)
        di = np.diag_indices(np.shape(t_mat)[0])
        t_mat[di] =  - self.ibd_out - self.ibd_jump + \
                       self.ibd_jump / (n)  # To account for jump to self
        t_mat[0, 0] = - self.ibd_in   # The rate of staying in diploid

        # Sanity Check if everything was filled correctly
        if (submat33 == False):
            assert(np.all(np.sum(t_mat, axis=1) > -0.0001))
            assert(np.all(np.sum(t_mat, axis=1) < 0.0001))
            #print(np.sum(t_mat, axis=1))  # Output

        self.trans_mat = t_mat
        return t_mat
    
    def full_transition_matrix(self, r_vec, t=[], n=4, submat33=True):
        """Compute and return the full transition Matrix.
        Calculates the first 3 states (not more needed by symmetry)
        t full Transition Matrix [k,k]. NO LOG STATE. If not given caluclate
        r_vec Map Length of Jumps [l] in Morgan
        n: Number of symmetric, non-background states"""
        ### infinitesimal rate
        if len(t)==0:
            t = self.calc_transition_rate(submat33=submat33, n=n)
            
        ### Full matrix
        r_vec = self.rmap_to_gaps(r_map=r_vec)
        if submat33:
            t = self.prep_3x3matrix(t, n)
        t_mat = self.exponentiate_r(t, r_vec)

        # Normalize to transition rate into non-collapsed state
        if submat33:
            t_mat[:, :2, 2] = t_mat[:, :2, 2] / (n - 1)
        return t_mat
    
    def prep_3x3matrix(self, t, n):
        """Prepare and return the grouped 3x3 Matrix 
        (grouped 3rd State: Everything in other IBD states)
        t: Origianl transition matrix (only first three entry important)
        n: Number of symmetric states
        """
        if self.output:
            print(f"HMM State Number: {n}")
        # Initiate to -1 (for later Sanity Check if everything is filled)
        t_simple = -np.ones((3, 3), dtype="float")
        t_simple[:2, :2] = t[:2, :2]
        # The probability of staying when in diff. ROH State:
        t_simple[2, 2] = -np.sum(t[2, :2])  # Minus the rates of Jumping
        # Jumping into 3rd state: Sum over all reference states
        t_simple[:2, 2] = t[:2, 2] * (n - 1) # Collapse the  other states
        t_simple[2, :2] = t[2, :2]  # The jumping out probability is the same
        return t_simple


    def exponentiate_r(self, rates, rec_v):
        """Calculates exponentiation of the rates matrix with rec_v
        rates: 2D Matrix of transitions
        rec_v: Array of length l"""
        eva, evec = np.linalg.eig(rates)  # Do the Eigenvalue Decomposition
        assert(np.max(eva) <= 1)   # Sanity Check whether rate Matrix
        evec_r = np.linalg.inv(evec)    # Do the Inversion
        # Create vector of the exponentiated diagonals
        d = np.exp(rec_v[:, None] * eva)
        # Use some Einstein Sum Convention Fun (C Speed):
        res = np.einsum('...ik, ...k, ...kj ->...ij', evec, d, evec_r)
        # Make sure that all transition rates are valuable
        assert(0 <= np.min(res))
        return res
    
    def rmap_to_gaps(self, r_map=[], cm=False):
        """Return the recombination map gaps [in Morgan]
        Input: Map Positions [l] (units see cm parameter below)
        Return: Rec. Distance Array [l]
        cm: Whether input is in centimorgan or morgan
        min_cap: Minimum Map Gap between Loci to cap
        max_cap: Maximum Mapg Gap between Loci to cap"""
        gaps = np.zeros(len(r_map))  # Make new array
        gaps[1:] = r_map[1:] - r_map[:-1]  # Calculate Differences
        assert(np.min(gaps) >= 0)
        if cm == True:
            gaps = gaps / 100     # Normalize to Morgan if map in cM
         
        ### Extend the minimum gap where needed
        gaps = np.maximum(gaps, self.min_gap)
        gaps = np.minimum(gaps, self.max_gap)

        if self.output == True:
            max_g = np.max(gaps)
            print(f"Minimum Genetic Map: {np.min(r_map):.4f} Morgan")
            print(f"Maximum Genetic Map: {np.max(r_map):.4f} Morgan")
            print(f"Gaps bigger than 0.1 cM: {np.sum(gaps > 0.001)}")
            print(f"Maximum Gap: {max_g * 100:.4f} cM")
            print(f"Upper Gap Cutoff: {self.max_gap * 100:.4f} cM")
        return gaps

class FiveStateAsymmetric(FiveStateTransitions):
    """A transition model for detecting IBD between ancient and modern sample pairs."""
    ibd_in = 1
    ibd_out = 20
    ibd_jump = 400 # IBD jump rate for ancient haplotypes
    ibd_jump2 = 0.5 # IBD jump rate for modern haplotypes

    def calc_transition_rate(self):
        """Return Transition Rate Matrix [k,k] to exponate. """
        t_mat = -np.ones((5, 5))

        t_mat[1:5, 0] = self.ibd_out  # The rate of jumping out IBD1 to nonIBD
        t_mat[0, 1:5] = self.ibd_in / 4  # Jumping into any IBD1 State from IBD0
        t_mat[1:5, 1:5] = self.ibd_jump / 4  # Jumping between IBD1 State
        t_mat[1, 2] = self.ibd_jump2
        t_mat[2, 1] = self.ibd_jump2
        t_mat[3, 4] = self.ibd_jump2
        t_mat[4, 3] = self.ibd_jump2

        # Do the Diagonal (do the usual model - for inf. substract 1)
        di = np.diag_indices(np.shape(t_mat)[0])
        t_mat[di] = -np.sum(t_mat, axis=1) + np.diag(t_mat)

        # Sanity Check if everything was filled correctly
        assert(np.allclose(np.sum(t_mat, axis=1), 0))
        self.trans_mat = t_mat
        return t_mat
    
    def full_transition_matrix(self, r_vec, t=[], n=4, submat33=True):
        """Compute and return the full transition Matrix.
        Calculates the first 3 states (not more needed by symmetry)
        t full Transition Matrix [k,k]. NO LOG STATE. If not given caluclate
        r_vec Map Length of Jumps [l] in Morgan
        n: Number of symmetric, non-background states, not used by this method, keeping it here to just keep function signature the same as that of the 5state HMM.
        submat33: also not used by this function, keeping it here just to keep consistency with the 5 state hmm
        Note: this is the only function called by outside code
        """
        ### infinitesimal rate
        if len(t) == 0:
            t = self.calc_transition_rate()
            
        ### Full matrix
        r_vec = self.rmap_to_gaps(r_map=r_vec)
        t_mat = self.exponentiate_r(t, r_vec)

        # Normalize to transition rate into non-collapsed state
        return t_mat


class SevenStateTransitions(FiveStateTransitions):
    """Implements the Model Transitions [units generally in Morgan]"""
    # We might want to consider a slower rate of jumping into IBD2 than jumping into IBD1, but let's leave it to be the same for now
    ibd_in = 1    # The rate of jumping into IBD copying state
    ibd_out = 20     # The rate of jumping out of IBD state
    ibd_jump = 500     # The rate of jumping within IBD to other haplotype pair
    #ibd_switch = 20  # the rate of jumping between IBD1 and IBD2 state, actually let's leave it the same as ibd_out for now
    

    def calc_transition_rate(self):
        """Return Transition Rate Matrix [k,k] to exponate.
        n: Number of symetric IBD states. Usually four (2x2 copying possibilities)
        submat33: Whether to only fill in the first 3 states """
        t_mat = -np.ones((7, 7))

        t_mat[1:5, 0] = self.ibd_out  # The rate of jumping out IBD1 to nonIBD
        t_mat[5:7, 0] = 1e-6 # the rate of jumping out IBD2 to nonIBD
        t_mat[0, 1:5] = self.ibd_in / 4  # Jumping into any IBD1 State from IBD0
        t_mat[0, 5:7] = 1e-6 # jumping into any IBD2 state from IBD0
        t_mat[1:5, 1:5] = self.ibd_jump / 4  # Jumping between IBD1 State
        t_mat[5:7, 5:7] = self.ibd_jump / 2  # Jumping between IBD2 State
        t_mat[1:5, 5:7] = 0.01 # jumping from IBD1 to IBD2 State
        t_mat[5:7, 1:5] = 200 # jumping from IBD2 to IBD1 state

        # Do the Diagonal (do the usual model - for inf. substract 1)
        di = np.diag_indices(np.shape(t_mat)[0])
        t_mat[di] = -np.sum(t_mat, axis=1) + np.diag(t_mat)

        # Sanity Check if everything was filled correctly
        assert(np.allclose(np.sum(t_mat, axis=1), 0))
        self.trans_mat = t_mat
        return t_mat
    
    def full_transition_matrix(self, r_vec, t=[], n=4, submat33=True):
        """Compute and return the full transition Matrix.
        Calculates the first 3 states (not more needed by symmetry)
        t full Transition Matrix [k,k]. NO LOG STATE. If not given caluclate
        r_vec Map Length of Jumps [l] in Morgan
        n: Number of symmetric, non-background states, not used by this method, keeping it here to just keep function signature the same as that of the 5state HMM.
        submat33: also not used by this function, keeping it here just to keep consistency with the 5 state hmm
        Note: this is the only function called by outside code
        """
        ### infinitesimal rate
        if len(t) == 0:
            t = self.calc_transition_rate()
            
        ### Full matrix
        r_vec = self.rmap_to_gaps(r_map=r_vec)
        t_mat = self.exponentiate_r(t, r_vec)

        # Normalize to transition rate into non-collapsed state
        return t_mat


############################################
# Factory method to return Transition Object

def load_transition_model(t_model="standard"):
    """Load the Transition Model"""
    if t_model == "standard":
        t_obj = FiveStateTransitions()
    elif t_model == "asymmetric":
        t_obj = FiveStateAsymmetric()
    elif t_model == 'IBD2':
        t_obj = SevenStateTransitions()
    else:
        raise NotImplementedError("Transition Model not found!")
    return t_obj



"""
Functions to detect duplicate Samples in imputed HDF5 file
@ Author: Harald Ringbauer, 2023
"""

import numpy as np
import pandas as pd
import h5py
import itertools as it

def get_match_df(path_h5="./data/hdf5/1240k_v54.1/ch", ch=3,
                 iids=[], gp_cutoff=0.98):
    """Create diploid Match Rate for all pairs in list of iids"""

    iids1, iids2, ms, os = [],[],[],[]

    with h5py.File(f"{path_h5}{ch}.h5", "r") as f:
        for iid1,iid2 in it.combinations(iids, r=2): 

            m,o = get_fraction_identical(f, sample1=iid1, sample2=iid2, 
                                   gp_cutoff=gp_cutoff, output=False)# Load for Sanity Check. See below!
            iids1.append(iid1)
            iids2.append(iid2)
            ms.append(m)
            os.append(o)

    dft = pd.DataFrame({"iid1":iids1,"iid2":iids2, "frac_match":ms, "frac_snps":os})
    return dft

def get_fraction_identical(f, sample1="SUC006", sample2="R26.SG", 
                           gp_cutoff=0.98, output=False):
    """Get Fraction of Identical Genotype Configurations.
    Return Fraction same IID, and fraction SNPs both IIDs tested"""
    if output:
        print(f"Running {sample1}-{sample2}")
    j1 = get_exact_idx_iid(f, sample1)[0]
    j2 = get_exact_idx_iid(f, sample2)[0]
    
    idx1 = get_markers_good(f, j1, cutoff=gp_cutoff, output=output)
    idx2 = get_markers_good(f, j2, cutoff=gp_cutoff, output=output)
    idx = (idx1 & idx2)
    snp_frac = np.mean(idx)
    if output:
        print(f"Filtering to common GP variants: {snp_frac:.3f}x")
    
    gt1 = f["calldata/GT"][:, j1, :][idx,:]
    gt2 = f["calldata/GT"][:, j2, :][idx,:]
    g1, g2 = np.sum(gt1, axis=1), np.sum(gt2, axis=1)
    frac_same = np.mean(g1 == g2)
    return frac_same,snp_frac

def get_exact_idx_iid(f, sample, unique=True):
    """Return Index of sample samples in hdf5 f"""
    samples = pd.Series(f["samples"][:].astype("str"))
    idx = samples==sample
    idx = np.where(idx)[0]
    assert(len(idx)==1)
    return idx

def get_markers_good(f, j, output=True, cutoff=0.99):
    """Get markers"""
    m = np.max(f["calldata/GP"][:,j,:], axis=1)
    idx = (m>cutoff)
    if output:
        c1 = np.mean(m>cutoff)
        print(f"Filtering to {cutoff} GP variants: {c1:.3f}x")
    return idx


"""
Functions to load data from a HDF5 file
@ Author: Harald Ringbauer, 2021
"""

import numpy as np
import pandas as pd
import os as os
import sys as sys
import h5py

def get_idx_iid(f,sample, unique=True):
    """Return Index of sample samples in hdf5 f"""
    samples = pd.Series(f["samples"].asstr()[:])
    idx = samples.str.contains(sample)
    idx = np.where(idx)[0]
    if unique:
        assert(len(idx)==1)
        return idx[0]
    else:
        return idx
    
def get_idx_iid_exact(f,sample, unique=True):
    """Return Index of sample samples in hdf5 f"""
    samples = pd.Series(f["samples"].asstr()[:])
    idx = samples==sample
    idx = np.where(idx)[0]
    if unique:
        assert(len(idx)==1)
        return idx[0]
    else:
        return idx
    
def get_coverage(f, j):
    """Get Coverage of sample j in hdf5 f"""
    ads =  f["calldata/AD"][:,j,:2]
    ads[ads<0]=0
    cov = np.mean(ads)*2
    return cov

def get_markers_good(f, j, output=True, cutoff=0.99):
    """Get markers"""
    m = np.max(f["calldata/GP"][:,j,:], axis=1)
    idx = (m>=cutoff)
    if output:
        c1 = np.mean(idx)
        print(f"Filtering to {cutoff} GP variants: {c1:.3f}x")
    return idx

###########################
### Load single individual

def get_genos(f, iid="SUC002", min_gp=0.98, output=True, phased=False, exact=True):
    """Return Genotypes and Map of pairs at intersection with GP>cutoff.
    phased: Whether to return [lx2] phased vector or [l] vetor of #derived.
    exact: Whether IID has to be an exact match"""
    if exact:
        j1 = get_idx_iid_exact(f, iid)  
    else:
        j1 = get_idx_iid(f, iid)
    
    m = f["variants/MAP"][:]
    pos = f["variants/POS"][:]
    gt1 = f["calldata/GT"][:, j1, :]
    gp1 = f["calldata/GP"][:, j1, :]    
    
    if min_gp>0:
        idx = get_markers_good(f, j1, cutoff=min_gp, output=output)
        m = m[idx]
        pos = pos[idx]
        gt1 = gt1[idx,:]
        gp1 = gp1[idx,:]
        
    if not phased:
        gt1 = np.sum(gt1, axis=1)
    return gt1, gp1, m, pos

def load_individual_h5(path_h5 = '/n/groups/reich/hringbauer/git/hapBLOCK/data/hdf5/1240k_v43/ch',
                       min_gp=0.98, chs = range(1,23), iid="SUC002", output=False):
    """Load individual data from set of chromosomal hdf5s."""
    dfs = []

    for ch in chs:
        path_h5_ch = f"{path_h5}{ch}.h5"
        with h5py.File(path_h5_ch, "r") as f: 
            gt1, gp1, m, pos = get_genos(f, iid=iid, min_gp=min_gp, output=False, phased=False) 
            gp_max = np.max(gp1, axis=1)
            dfs.append(pd.DataFrame({"ch":ch, "pos":pos, "gt":gt1, "gp":gp_max}))

    df_ind = pd.concat(dfs)
    return df_ind

#############################
### Extract Pairs of Markers

def get_genos_pairs(f, sample1="SUC006", sample2="R26.SG", 
                    cutoff=0.98, output=True, phased=False, exact=False):
    """Return Genotypes and Map of pairs at intersection with GP>cutoff.
    phased: Whether to return [lx2] phased vector or [l] vetor of #derived.
    exact: Whether IID has to be an exact match"""
    if exact:
        j1 = get_idx_iid_exact(f, sample1)
        j2 = get_idx_iid_exact(f, sample2)
        
    else:
        j1 = get_idx_iid(f, sample1)
        j2 = get_idx_iid(f, sample2)
    
    idx1 = get_markers_good(f, j1, cutoff=cutoff, output=output)
    idx2 = get_markers_good(f, j2, cutoff=cutoff, output=output)
    idx = (idx1 & idx2)
    snp_frac = np.mean(idx)
    if output:
        print(f"Filtering to common GP variants: {snp_frac:.3f}x")
    
    m = f["variants/MAP"][:][idx]
    
    gt1 = f["calldata/GT"][:, j1, :][idx,:]
    gt2 = f["calldata/GT"][:, j2, :][idx,:]
    if not phased:
        gt1, gt2 = np.sum(gt1, axis=1), np.sum(gt2, axis=1)
    return gt1, gt2, m

def opp_homos(g1, g2):
    """Return opposing homozygotes"""
    o1 = (g1 == 0) & (g2 == 2)
    o2 = (g1 == 2) & (g2 == 0)
    return (o1 | o2)

def get_opp_homos_f(f_path="/n/groups/reich/hringbauer/git/hapBLOCK/data//hdf5/1240k_v43/ch",
                    iid1="SUC006", iid2="R26.SG", ch=3,
                    cutoff=0.99, output=True, exact=False):
    """Return opposing homozygotes boolean array and map array at intersection with
    GP>cutoff."""
    load_path = f_path + str(ch) + ".h5"
    with h5py.File(load_path, "r") as f:
        g1, g2, m = get_genos_pairs(f, sample1=iid1, sample2=iid2, 
                                cutoff=cutoff, output=output, exact=exact)
        o_homos = opp_homos(g1, g2)
    return o_homos, m


# cython: language_level=3, boundscheck=True, wraparound=False
import numpy as np
import psutil      # For Memory Profiling
import os
cimport cython

#from scipy.special import logsumexp
from libc.math cimport exp, log   # For doing Exponentials and Logs

DTYPE = np.float64 # The float data type


cdef inline double logsumexp(double[:] vec):
  """Do the Log of the Sum of Exponentials."""
  cdef Py_ssize_t i  # The iterator Variable
  cdef double result = 0.0
  cdef double largest = vec[0]

  for i in range(1, vec.shape[0]):   # Find Maximum in vec
      if (vec[i] > largest):
          largest = vec[i]
  for i in range(vec.shape[0]):
      result += exp(vec[i] - largest)
  return largest + log(result)

cdef inline long argmax(double[:] vec):
  """Return Max and ArgMax"""
  cdef Py_ssize_t i  # The iterator Variable.
  cdef int m = 0     # Position of the Maximum.
  cdef double v = vec[0]

  for k in range(1, vec.shape[0]):   # Find Maximum
    if (vec[k] > v):
      m, v = k, vec[k]
  return m  # Return Argmax

cdef inline double sum_array(double[:] vec, int n):
    """Sum over array.
    vec: Array to sum
    n: Number of elements"""
    cdef double s = vec[0]
    for i in range(1,n):
        s = s + vec[i]
    return s

def print_memory_usage():
    """Print the current Memory Usage in mB"""
    process = psutil.Process(os.getpid())
    mb_usage = process.memory_info().rss / 1e6
    print(f"Memory Usage: {mb_usage} mB")


def fwd_bkwd(double[:, :] e_mat, double[:, :] t_mat,
             double[:, :] fwd, double[:, :] bwd, double[:,:,:] t, 
             full=False, output=True):
    """Takes emission and transition probabilities, and calculates posteriors.
    Input: kxl matrices of emission, transition
    and initialized fwd and bwd probabilities. Given in normal space
    full: Boolean whether to return everything.
    output: Whether to print output useful for monitoring"""
    cdef int n_states = e_mat.shape[0]
    cdef int n_loci = e_mat.shape[1]
    cdef Py_ssize_t i, j, k # The Array Indices

    # Initialize Posterior and Transition Probabilities
    post = np.empty([n_states, n_loci], dtype=DTYPE)

    trans_ll = np.empty(n_states, dtype=DTYPE)
    cdef double[:] trans_ll_view = trans_ll

    ### Transform to Log space
    cdef double[:, :] t_mat0 = np.log(np.eye(n_states) + t_mat[:,:])  # Do log of (relevant) transition Matrix
    cdef double[:, :] e_mat0 = np.log(e_mat)
    

    for i in range(1, n_loci):  # Do the forward recursion
        for j in range(n_states):
          for k in range(n_states):
            trans_ll_view[k] = fwd[k, i - 1] + t_mat0[k, j]

          fwd[j, i] = e_mat0[j, i] + logsumexp(trans_ll_view)

    for i in range(n_loci-1, 0, -1):  # Do the backward recursion
      for j in range(n_states):
        for k in range(n_states):
          trans_ll_view[k] = t_mat0[j, k] + e_mat0[k, i] + bwd[k, i]
        bwd[j, i - 1] = logsumexp(trans_ll_view)

    # Get total log likelihood
    for k in range(n_states):  # Simply sum the two 1D arrays
      trans_ll_view[k] = fwd[k, n_loci-1] + bwd[k, n_loci-1]
    tot_ll = logsumexp(trans_ll_view)
    if output:
        print(f"Total Log likelihood: {tot_ll: .3f}")

    # Combine the forward and backward calculations
    fwd1 = np.asarray(fwd)  # Transform
    bwd1 = np.asarray(bwd)
    post = fwd1 + bwd1 - tot_ll
    post = np.exp(post) ## Transform from log to normal space

    if full==False:
      return post

    elif full==True:   # Return everything
      return post, fwd1, bwd1, tot_ll


def fwd_bkwd_fast(double[:, :] e_mat, double[:, :, :] t_mat, double in_val = 1e-4, 
                  full=False, output=True):
    """Takes emission and transition probabilities, and calculates posteriors.
    Uses speed-up specific for Genotype data (pooling same transition rates)
    Input:
    e_mat: Emission probabilities [k x l] (normal space)
    t_mat: Transition Matrix: [l x 3 x 3]  (normal space)
    in_val: Intitial probability of single symmetric state (normal space)
    full: Boolean whether to return everything (post, fwd1, bwd1, tot_ll)
    output: Whether to print output useful for monitoring
    Otherwise only posterior mat [kxl] of post is returned
    """
    cdef int n_states = e_mat.shape[0]
    cdef int n_loci = e_mat.shape[1]
    cdef Py_ssize_t i, j, k    # The Array Indices
    cdef double stay           # The Probablility of Staying

    # Initialize Posterior and Transition Probabilities
    post = np.empty([n_states, n_loci], dtype=DTYPE)

    trans_ll = np.empty(n_states-1, dtype=DTYPE) # Array for pre-calculations
    cdef double[:] trans_ll_view = trans_ll

    trans_ll1 = np.empty(n_states, dtype=DTYPE) # Array for calculations
    cdef double[:] trans_ll_view1 = trans_ll1

    three_v = np.empty(3, dtype=DTYPE)     # Array of size three
    cdef double[:] three_v_view = three_v

    two_v = np.empty(2, dtype=DTYPE)       # Array of size two
    cdef double[:] two_v_view = two_v

    # Do transform to Log Space:
    cdef double[:,:,:] t0 = np.log(t_mat)         # Do log of recombination Map
    cdef double[:, :] e_mat0 = np.log(e_mat)

    ### Initialize FWD BWD matrices
    fwd0 = np.zeros((n_states, n_loci), dtype="float")
    fwd0[:, 0] = np.log(in_val)  # Initial Probabilities
    fwd0[0, 0] = np.log(1 - (n_states - 1) * in_val)
    cdef double[:,:] fwd = fwd0

    bwd0 = np.zeros((n_states, n_loci), dtype="float")
    bwd0[:, -1] = np.log(in_val)
    bwd0[0, -1] = np.log(1 - (n_states - 1) * in_val)
    cdef double[:,:] bwd = bwd0

    #############################
    ### Do the Forward Algorithm
    for i in range(1, n_loci):  # Run forward recursion
        stay = log(t_mat[i, 1, 1] - t_mat[i, 1, 2])  # Do the log of the Stay term

        for k in range(1, n_states): # Calculate logsum of ROH states:
            trans_ll_view[k-1] = fwd[k, i - 1]
        f_l = logsumexp(trans_ll_view) # Logsum of ROH States

        # Do the 0 State:
        two_v_view[0] = fwd[0, i - 1] + t0[i, 0, 0]   # Staying in 0 State
        two_v_view[1] = f_l + t0[i, 1, 0]             # Going into 0 State
        fwd[0, i] = e_mat0[0, i] + logsumexp(two_v_view)

        ### Do the other states
        # Preprocessing:
        three_v_view[0] = fwd[0, i - 1] + t0[i, 0, 1]   # Coming from 0 State
        three_v_view[1] = f_l + t0[i, 1, 2]             # Coming from other ROH State

        for j in range(1, n_states):  # Do the final run over all states
          three_v_view[2] = fwd[j, i-1] +  stay
          fwd[j, i] = e_mat0[j, i] + logsumexp(three_v_view)

    #############################
    ### Do the Backward Algorithm
    for i in range(n_loci-1, 0, -1):  # Run backward recursion
      stay = log(t_mat[i, 1, 1] - t_mat[i, 1, 2])

      for k in range(1, n_states): # Calculate logsum of ROH states:
          trans_ll_view[k-1] = bwd[k, i] + e_mat0[k, i]
      f_l = logsumexp(trans_ll_view) # Logsum of ROH States

      # Do the 0 State:
      two_v_view[0] = bwd[0, i] + t0[i, 0, 0] + e_mat0[0, i]   # Staying in 0 State
      two_v_view[1] = f_l + t0[i, 0, 1]                         # Going into 0 State
      bwd[0, i - 1] = logsumexp(two_v_view)

      ### Do the other states
      # Preprocessing:
      three_v_view[0] = e_mat0[0, i] + bwd[0, i] + t0[i, 1, 0]
      three_v_view[1] = f_l + t0[i, 1, 2]    # Coming from other ROH State

      for j in range(1, n_states):  # Do the final run over all states
        three_v_view[2] = e_mat0[j, i] + bwd[j, i] +  stay
        bwd[j, i - 1] = logsumexp(three_v_view)  # Fill in the backward Probability

    # Get total log likelihood
    for k in range(n_states):  # Simply sum the two 1D arrays
      trans_ll_view1[k] = fwd[k, n_loci - 1] + bwd[k, n_loci - 1]
    tot_ll = logsumexp(trans_ll_view1)

    # Combine the forward and backward calculations
    fwd1 = np.asarray(fwd, dtype=DTYPE)  ## Transform
    bwd1 = np.asarray(bwd, dtype=DTYPE)
    post = fwd1 + bwd1 - float(tot_ll) ## Posterior in log space
    post = np.exp(post)        ## To have posterior in normal space
    if output:
        print("Memory Usage Full:")
        print_memory_usage()   ## For MEMORY_BENCH
        print(f"Total Log likelihood: {tot_ll: .3f}")

    if full==False:
      return post

    elif full==True:   # Return everything
      return post, fwd1, bwd1, tot_ll


def fwd_bkwd_lowmem(double[:, :] e_mat, double[:, :, :] t_mat, double in_val = 1e-4, 
                    full=False, output=True):
    """Takes emission and transition probabilities, and calculates posteriors.
    Uses speed-up specific for Genotype data (pooling same transition rates)
    Low-Mem: Do no save the full FWD BWD and Posterior. Use temporary
    Arrays for saving. Only return 0-posterior. useful for large symmetric state arrays.
    Input:
    e_mat0: Emission probabilities [k x l] (normal space)
    t_mat: Transition Matrix: [l x 3 x 3]  (normal space)
    in_val: Intitial probability of single symmetric state (normal space)
    full: Boolean whether to return everything (post, fwd1, bwd1, tot_ll)
    Otherwise only posterior mat [kxl] of post is returned"""
    cdef int n_states = e_mat.shape[0]
    cdef int n_loci = e_mat.shape[1]
    cdef Py_ssize_t i, j, k    # The Array Indices
    cdef double stay           # The Probablility of Staying
    cdef double tot_ll  # The total Likelihood (need for Posterior)

    # Initialize Posterior and Transition Probabilities
    post = np.empty(n_loci, dtype=DTYPE) # Array of 0 State Posterior
    cdef double[:] post_view = post

    trans_ll = np.empty(n_states-1, dtype=DTYPE) # Array for pre-calculations
    cdef double[:] trans_ll_view = trans_ll

    trans_ll1 = np.empty(n_states, dtype=DTYPE) # Array for calculations
    cdef double[:] trans_ll_view1 = trans_ll1

    three_v = np.empty(3, dtype=DTYPE)     # Array of size three
    cdef double[:] three_v_view = three_v

    two_v = np.empty(2, dtype=DTYPE)       # Array of size two
    cdef double[:] two_v_view = two_v

    # Do transform to Log Space:
    cdef double[:,:,:] t0 = np.log(t_mat) 
    cdef double[:, :] e_mat0 = np.log(e_mat)

    ### Initialize FWD BWD Arrays
    fwd0 = np.zeros(n_states, dtype=DTYPE)
    fwd0[:] = np.log(in_val)  # Initial Probabilities
    fwd0[0] = np.log(1 - (n_states - 1) * in_val)
    cdef double[:] fwd = fwd0

    bwd0 = np.zeros(n_states, dtype=DTYPE)
    bwd0[:] = np.log(in_val)
    bwd0[0] = np.log(1 - (n_states - 1) * in_val)
    cdef double[:] bwd = bwd0

    tmp0 = np.zeros(n_states, dtype=DTYPE)
    cdef double[:] tmp = tmp0

    #############################
    ### Do the Forward Algorithm

    post_view[0] = fwd[0] # Add to first locus 0 Posterior
    for i in range(1, n_loci):  # Run forward recursion
        stay = log(t_mat[i, 1, 1] - t_mat[i, 1, 2])  # Do the log of the Stay term

        for k in range(1, n_states): # Calculate logsum of ROH states:
            trans_ll_view[k-1] = fwd[k]
        f_l = logsumexp(trans_ll_view) # Logsum of ROH States

        # Do the 0 State:
        two_v_view[0] = fwd[0] + t0[i, 0, 0]   # Staying in 0 State
        two_v_view[1] = f_l + t0[i, 1, 0]             # Going into 0 State
        tmp[0] = e_mat0[0, i] + logsumexp(two_v_view)

        ### Do the other states
        # Preprocessing:
        three_v_view[0] = fwd[0] + t0[i, 0, 1]   # Coming from 0 State
        three_v_view[1] = f_l + t0[i, 1, 2]             # Coming from other ROH State

        for j in range(1, n_states):  # Do the final run over all states
          three_v_view[2] = fwd[j] +  stay
          tmp[j] = e_mat0[j, i] + logsumexp(three_v_view)

        ### Make tmp new fwd vec:
        for j in range(0, n_states):
          fwd[j] = tmp[j]
        post_view[i] = fwd[0]  # Add to 0-State Posterior

    ### Get total log likelihood
    for k in range(n_states):  # Simply sum the two 1D arrays
      trans_ll_view1[k] = fwd[k] + bwd[k]
    tot_ll = logsumexp(trans_ll_view1)

    #############################
    ### Do the Backward Algorithm
    ## last0-State Posterior
    post_view[n_loci-1] = post_view[n_loci-1] + bwd[0] - tot_ll

    for i in range(n_loci-1, 0, -1):  # Run backward recursion
      stay = log(t_mat[i, 1, 1] - t_mat[i, 1, 2])

      for k in range(1, n_states): # Calculate logsum of ROH states:
          trans_ll_view[k-1] = bwd[k] + e_mat0[k, i]
      f_l = logsumexp(trans_ll_view) # Logsum of ROH States

      # Do the 0 State:
      two_v_view[0] = bwd[0] + t0[i, 0, 0] + e_mat0[0, i]   # Staying in 0 State
      two_v_view[1] = f_l + t0[i, 0, 1]                         # Going into 0 State
      tmp[0] = logsumexp(two_v_view)

      ### Do the other states
      # Preprocessing:
      three_v_view[0] = e_mat0[0, i] + bwd[0] + t0[i, 1, 0]
      three_v_view[1] = f_l + t0[i, 1, 2]    # Coming from other ROH State

      for j in range(1, n_states):  # Do the final run over all states
        three_v_view[2] = e_mat0[j, i] + bwd[j] +  stay
        tmp[j] = logsumexp(three_v_view)  # Fill in the backward Probability

      ### Make tmp new bwd vec:
      for j in range(0, n_states):
        bwd[j] = tmp[j]

      ### Finalize the 0 Posterior
      post_view[i-1] = post_view[i-1] + bwd[0] - tot_ll
    
    post = np.exp(post) # To transform to normal space
    if output:
        print(f"Total Log likelihood: {tot_ll: .3f}")
        print_memory_usage()   ## For MEMORY_BENCH
    
    if full==False:
      return post[None,:]  # For "fake" axis

    elif full==True:   # Return everything
      return post[None,:], fwd0, bwd0, tot_ll

###############################################################################

def fwd_bkwd_scaled(double[:, :] e_mat, double[:, :, :] t_mat, 
                    double in_val = 1e-4, full=False, output=True):
    """Takes emission and transition probabilities, and calculates posteriors.
    Uses speed-up specific for Genotype data (pooling same transition rates)
    Uses rescaling of fwd and bwd matrices
    e_mat: Emission probabilities [k x l] (normal space)
    t_mat: Transition Matrix: [l x 3 x 3]  (normal space)
    in_val: Intitial probability of single symmetric state (normal space)
    full: Boolean whether to return everything (post, fwd1, bwd1, tot_ll)
    output: Whether to print output useful for monitoring
    Otherwise only posterior mat [kxl] of post is returned
    """
    cdef int n_states = e_mat.shape[0]
    cdef int n_loci = e_mat.shape[1]
    cdef Py_ssize_t i, j, k    # The Array and Iteration Indices
    cdef double stay           # The Probablility of Staying
    cdef double x1, x2, x3     # Place holder variables [make code readable]

    # Initialize Posterior and Transition Probabilities
    post = np.empty([n_states, n_loci], dtype=DTYPE)
    
    c = np.empty(n_loci, dtype=DTYPE) # Array of normalization constants
    cdef double[:] c_view = c
    
    temp = np.empty(n_states, dtype=DTYPE) # l Array for calculations
    cdef double[:] temp_v = temp
    
    temp1 = np.empty(n_states-1, dtype=DTYPE) # l-1 Array for calculatons
    cdef double[:] temp1_v = temp1

    cdef double[:,:,:] t = t_mat   # C View of transition matrix

    ### Initialize FWD BWD matrices with first and last entries filled
    fwd1 = np.zeros((n_states, n_loci), dtype="float")
    fwd1[:, 0] = in_val  # Initial Probabilities
    fwd1[0, 0] = 1 - (n_states - 1) * in_val
    cdef double[:,:] fwd = fwd1
    
    c_view[0] = 1 # Set the first normalization constant

    bwd1 = np.zeros((n_states, n_loci), dtype="float")
    bwd1[:, -1] = 1 # The initial values for the backward pass
    #bwd1[0, -1] = 1 - (n_states - 1) * in_val
    cdef double[:,:] bwd = bwd1

    #############################
    ### Do the Forward Algorithm
    for i in range(1, n_loci):  # Run forward recursion
        stay = t[i, 1, 1] - t[i, 1, 2]  # Do the log of the Stay term

        #for k in range(1, n_states): # Calculate Sum of ROH states. 
        f_l = 1 - fwd[0, i-1]  ### Assume they are normalized!!!
        
        ### Do the 0 State:
        x1 = fwd[0, i - 1] * t[i, 0, 0]    # Staying in 0 State
        x2 = f_l * t[i, 1, 0]               # Going into 0 State from any other
        temp_v[0] = e_mat[0, i] * (x1 + x2) # Set the unnorm. 0 forward variable

        ### Do the other states
        # Preprocessing:
        x1 = fwd[0, i - 1] * t[i, 0, 1]   # Coming from 0 State
        x2 = f_l * t[i, 1, 2]             # Coming from other ROH State

        for j in range(1, n_states):  # Do the final run over all states
            x3 = fwd[j, i-1] *  stay # Staying in state
            temp_v[j] = e_mat[j, i] * (x1 + x2 + x3)
            
        ### Do the normalization
        c_view[i] = sum_array(temp_v, n_states)
        for j in range(n_states):
            fwd[j,i] = temp_v[j] / c_view[i] # Rescale to prob. distribution
            
    #############################
    ### Do the Backward Algorithm
    for i in range(n_loci-1, 0, -1):  # Run backward recursion
        stay = t[i, 1, 1] - t[i, 1, 2]

        for k in range(1, n_states): # Calculate logsum of ROH states:
            temp1_v[k-1] = bwd[k, i] * e_mat[k, i]
        f_l = sum_array(temp1_v, n_states-1) # Logsum of ROH States

      # Do the 0 State:
        x1 = bwd[0, i] * t[i, 0, 0] * e_mat[0, i]   # Staying in 0 State
        x2 = f_l * t[i, 0, 1]                         # Going into 0 State
        temp_v[0] = x1 + x2

      ### Do the other states
      # Preprocessing:
        x1 = e_mat[0, i] * bwd[0, i] * t[i, 1, 0]
        x2 = f_l * t[i, 1, 2]    # Coming from other ROH State

        for j in range(1, n_states):  # Do the final run over all states
            x3 = e_mat[j, i] * bwd[j, i] *  stay
            temp_v[j] = x1 + x2 + x3  # Fill in the backward Probability
        
        ### Do the normalization
        for j in range(n_states):
            bwd[j, i - 1] = temp_v[j] / c_view[i] # Rescale to prob. distribution
    
    ### Combine the forward and backward calculations for posterior
    post = fwd1 * bwd1
    if output:
        print("Memory Usage at end of HMM:")
        print_memory_usage()   ## For MEMORY_BENCH

    if full==False:
        return post

    elif full==True:   # Return everything
        tot_ll = np.sum(np.log(c)) # Tot Likelihood is product over all c. 
        if output:
            print(f"Total Log likelihood: {tot_ll: .3f}")
        return post, fwd1, bwd1, tot_ll

###############################################################################
###############################################################################
### Viterbi Path

def viterbi_path(double[:, :] e_mat0, double[:, :, :] t_mat0, double[:] end_p0):
    """Implementation of a Viterbi Path.
    e_mat0  Matrices with Emission Probabilities, [k,l] (log space)
    t_mat: Transition Matrix: [l x 3 x 3]  (normal space)
    end_p: probability to begin/end in states [k]"""
    cdef int n_states = e_mat0.shape[0]
    cdef int n_loci = e_mat0.shape[1]
    cdef Py_ssize_t i, j, k # The Array Indices
    cdef int m  # Placeholder for Maximum
    cdef double v # Value to set

    # Do the actual optimization (with back-tracking)
    # Initialize Views:
    cdef double[:, :] mp = np.empty((n_states, n_loci), dtype=DTYPE)
    cdef double[:] new_p = np.empty(n_states, dtype = DTYPE) # Temporary Array
    cdef long[:,:] pt = np.empty((n_states, n_loci), dtype = np.int16)  # Previous State Pointer

    trans_ll = np.empty(n_states-1, dtype=DTYPE) # Array for pre-calculations
    cdef double[:] trans_ll_view = trans_ll

    three_v = np.empty(3, dtype=DTYPE)     # Array of size three
    cdef double[:] three_v_view = three_v

    three_vi = np.empty(3, dtype=int)       # Int Array of size three
    cdef long[:] three_vi_view = three_vi

    two_v = np.empty(2, dtype=DTYPE)       # Array of size two
    cdef double[:] two_v_view = two_v

    two_vi = np.empty(2, dtype=int)       # Int Array of size two
    cdef long[:] two_vi_view = two_vi

    for k in range(n_states):
      mp[k, 0] = end_p0[k]  # Initialize with Ending Probabilities

      two_vi_view[0] = 0
      three_vi_view[0] = 0

    for i in range(1, n_loci):  # Do the Viterbi-Iteration
        ### Precomputation:
        # Do the maximal log probability of 1, ...k State:
        m = argmax(mp[1:, i - 1])
        v = mp[m+1, i - 1]

        # Do the States from collapsed states
        two_vi_view[1] = m+1   # Set the Pointers
        three_vi_view[1] = m+1

        two_v_view[1] = v + t_mat0[i, 1, 0]
        three_v_view[1] = v + t_mat0[i, 1, 2] # Move in from other ROH

        ### Do the zero State
        two_v_view[0] = mp[0, i - 1] + t_mat0[i, 0, 0]

        m = argmax(two_v_view)      ### Do a Maximum
        v = two_v_view[m]
        mp[0, i] = v + e_mat0[0, i]   ### Set Max. Probability
        pt[0, i] = two_vi_view[m]      ### Set Pointer for Backtrace

        ### Do the other States
        three_v_view[0] = mp[0, i - 1] + t_mat0[i, 0, 1] # Move from 0 State

        for k in range(1, n_states):   # Find Maximum
          three_v_view[2] = mp[k, i - 1] + t_mat0[i, 1, 1] # The Stay State
          three_vi_view[2] = k

          m = argmax(three_v_view)      ### Do a Maximum
          v = three_v_view[m]

          mp[k, i] = v + e_mat0[k, i]   ### Set Max. Probability
          pt[k, i] = three_vi_view[m]      ### Set Pointer for Backtrace

    ### Do the trace back
    cdef long[:] path = -np.ones(n_loci, dtype=np.int16)  # Initialize

    x = np.argmax(mp[:, n_loci-1])  # The highest probability
    path[n_loci-1] = x

    for i in range(n_loci - 1, 0, -1):
        # Always th pointer to the previous path
        path[i - 1] = pt[path[i], i]

    m = path[n_loci-1]
    print(f"Log likelihood Path: {mp[m,n_loci-1]:.3f}")
    assert(np.min(path)>=0) #Sanity check if everything was filled up
    return np.asarray(path)


################################################### code for IBD2 states ###########################################

cdef inline void elementwise_multiple(double[:] vec1, double[:] vec2, double[:] dest, int n):
  for i in range(n):
    dest[i] = vec1[i]*vec2[i]

cdef inline void set2zero(double[:] vec, int n):
  for i in range(n):
    vec[i] = 0.0


def fwd_bkwd_scaled_asymmetricTransition(double[:, :] e_mat, double[:, :, :] t_mat, 
                    double in_val = 1e-4, full=False, output=True):
    """Takes emission and transition probabilities, and calculates posteriors.
    Uses speed-up specific for Genotype data (pooling same transition rates)
    Uses rescaling of fwd and bwd matrices
    e_mat: Emission probabilities [k x l] (normal space)
    t_mat: Transition Matrix: [l x 7 x 7]  (normal space)
    in_val: Intitial probability of single symmetric state (normal space)
    full: Boolean whether to return everything (post, fwd1, bwd1, tot_ll)
    output: Whether to print output useful for monitoring
    Otherwise only posterior mat [kxl] of post is returned
    """
    cdef int n_states = e_mat.shape[0]
    cdef int n_loci = e_mat.shape[1]
    cdef Py_ssize_t i, j, k    # The Array and Iteration Indices
    cdef double stay           # The Probablility of Staying
    cdef double x1, x2, x3     # Place holder variables [make code readable]

    # Initialize Posterior and Transition Probabilities
    post = np.empty([n_states, n_loci], dtype=DTYPE)
    
    c = np.empty(n_loci, dtype=DTYPE) # Array of normalization constants
    cdef double[:] c_view = c
    
    temp = np.empty(n_states, dtype=DTYPE) # l Array for calculations
    cdef double[:] temp_v = temp
    
    temp1 = np.empty(n_states-1, dtype=DTYPE) # l-1 Array for calculatons
    cdef double[:] temp1_v = temp1

    temp2 = np.empty(n_states, dtype=DTYPE) # l Array for temporary storage (added by Yilei)
    cdef double[:] temp2_v = temp2

    cdef double[:,:,:] t = t_mat   # C View of transition matrix

    ### Initialize FWD BWD matrices with first and last entries filled
    fwd1 = np.zeros((n_states, n_loci), dtype="float")
    fwd1[:, 0] = in_val  # Initial Probabilities
    fwd1[0, 0] = 1 - (n_states - 1) * in_val
    cdef double[:,:] fwd = fwd1
    
    c_view[0] = 1 # Set the first normalization constant

    bwd1 = np.zeros((n_states, n_loci), dtype="float")
    bwd1[:, -1] = 1 # The initial values for the backward pass
    cdef double[:,:] bwd = bwd1

    #############################
    ### Do the Forward Algorithm
    for i in range(1, n_loci):  # Run forward recursion
      set2zero(temp_v, n_states)
      for j in range(n_states):
        for k in range(n_states):
          temp_v[j] += fwd[k, i-1]*t[i, k, j]
        temp_v[j] *= e_mat[j, i]
        
        ### Do the normalization
        c_view[i] = sum_array(temp_v, n_states)
        for j in range(n_states):
            fwd[j,i] = temp_v[j] / c_view[i] # Rescale to prob. distribution
            
    #############################
    ### Do the Backward Algorithm
    for i in range(n_loci-1, 0, -1):  # Run backward recursion
      set2zero(temp_v, n_states)
      for j in range(n_states):
        for k in range(n_states):
          temp_v[j] += bwd[k, i]*t[i, j, k]*e_mat[k, i]
      
      ### Do the normalization
      for j in range(n_states):
        bwd[j, i - 1] = temp_v[j] / c_view[i] # Rescale to prob. distribution
    
    ### Combine the forward and backward calculations for posterior
    post = fwd1 * bwd1
    if output:
        print("Memory Usage at end of HMM:")
        print_memory_usage()   ## For MEMORY_BENCH

    if full==False:
        return post

    elif full==True:   # Return everything
        tot_ll = np.sum(np.log(c)) # Tot Likelihood is product over all c. 
        if output:
            print(f"Total Log likelihood: {tot_ll: .3f}")
        return post, fwd1, bwd1, tot_ll



"""
Methods for fwd_bwd calculations. With Factory methods.
Be careful with c functions, they need exact type input.
Contains Sub-Classes, as well as factory Method.
@ Author: Harald Ringbauer, 2019, All rights reserved
"""
import numpy as np
from scipy.special import logsumexp as logsumexp
import sys
from ancIBD.cfunc import fwd_bkwd_fast, fwd_bkwd_lowmem, fwd_bkwd_scaled, fwd_bkwd_scaled_asymmetricTransition
#from ancIBD.cfunc import fwd_bkwd_fast, fwd_bkwd_lowmem, fwd_bkwd_scaled, fwd_bkwd_scaled_7States

####################################################
### Python function FWD BWD. (C FUNCTIONS GET IMPORTED)

def fwd_bkwd_p(e_prob, t_mat,  in_val = 1e-4, 
               full=True, output=True):
    """Simple Python Implementation of forward / bacdkward to compare 
    optimized c versions to. Uses standard definitions.
    Takes emission and transition probabilities
    Input: 
    Emission Matrix e_prob0 [l,k] NORMAL SPACE
    Transition Matrix t_mat [l,k,k] NORMALIZED
    and initialized fwd and bwd probabilities. All in log Space
    output: Whether to print out output useful for debugging/monitoring
    Output: post, fwd0, bwd0, tot_ll; if full then only post"""
    e_prob0 = np.log(e_prob)
    n_states, n_loci = np.shape(e_prob0)
    
    ### Initialize FWD BWD and Posterior
    fwd0 = np.zeros((n_states, n_loci), dtype="float")
    t_mat0 = np.log(t_mat)
    
    fwd0[:, 0] = np.log(in_val)  # Initial Probabilities
    fwd0[0, 0] = np.log(1 - (n_states - 1) * in_val)

    bwd0 = np.zeros((n_states, n_loci), dtype="float")
    bwd0[:, -1] = np.log(in_val)
    bwd0[0, -1] = np.log(1 - (n_states - 1) * in_val)
    
    ### Forward Run
    for i in range(1, n_loci): 
        for j in range(n_states):
            trans_ll = fwd0[:, i - 1] + t_mat0[i, :, j]
            fwd0[j, i] = e_prob0[j, i] + logsumexp(trans_ll)
    
    ### Backward Run
    for i in range(n_loci - 1, 0, -1):  # Do the backward recursion
        for j in range(n_states):
            trans_ll = t_mat0[i, j, :] + e_prob0[:, i] + bwd0[:, i]
            bwd0[j, i - 1] = logsumexp(trans_ll)

    tot_ll = logsumexp(fwd0[:, -1] + bwd0[:, -1])  # Get total log likelihood from last entry
    
    if output:
        print(f"Total Log likelihood fwd: {tot_ll: .3f}")
    # Combine the forward and backward calculations
    post = fwd0 + bwd0 - tot_ll  # The formulat is f*b/tot_l
    
    if full:
        return post, fwd0, bwd0, tot_ll
    else:
        return post        

####################################################
####################################################
### Additional Helper Functions

def print_memory_usage():
    """Print the current Memory Usage in mB"""
    process = psutil.Process(os.getpid())
    mb_usage = process.memory_info().rss / 1e6
    print(f"Memory Usage: {mb_usage} mB")
    
    
####################################################
### Factory method to load the right function

def load_fwd_bwd_func(h_model="FiveState"):
    """Return fwd_bwd function"""
    if h_model == "FiveState":
        func = fwd_bkwd_p
    elif h_model == "FiveStateFast":
        func = fwd_bkwd_fast
    elif h_model == "FiveStatLowMem":
        func = fwd_bkwd_lowmem
    elif h_model == "FiveStateScaled":
        func = fwd_bkwd_scaled
    elif h_model == 'asymmetric':
        func = fwd_bkwd_scaled_asymmetricTransition
    else:
        raise NotImplementedError("HMM Model not found!")
    return func


"""
Functions to QC imputed data (in hdf5 format)
@ Author: Harald Ringbauer, 2021
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os as os
import sys as sys
import h5py

def get_gp_df(path_h5 = "", chs = [1,], iids = ["iid1",], 
              cutoffs = [0.5,0.6,0.7,0.8,0.9]):
    """Check Genotype Probabilities for list of [iids] and list of [chs].
    Calculate max gp fractions <[cutoffs]. A value >0.01 for c=0.5 seems
    highly problematic, and should be flagged!
    Return summary dataframe for each chromosome and iid.
    path_h5: Path of the hdf5 up to the chr. number."""

    columns = ["iid", "ch", "n"] + cutoffs
    df = pd.DataFrame(columns=columns)

    for ch in chs:
        ### load the data
        print(f"Loading Chromosome {ch}...")
        with h5py.File(f"{path_h5}{ch}.h5", "r") as f: # Load for Sanity Check. See below!
                gt = f["calldata/GT"][:]
                gp = f["calldata/GP"][:]
                samples = f["samples"][:].astype("str")
                cm = f["variants/MAP"][:]

        ### iterate over iids
        for iid in iids:
            idx = (samples==iid)
            assert(np.sum(idx)==1)
            i = np.where(idx)[0][0]
            y = np.max(gp[:,i,:], axis=1)
            vs = [np.mean(y<=c) for c in cutoffs] # Values to append

            ### Append Values
            to_append = [iid, ch, len(y)] + vs
            a_series = pd.Series(to_append, index = df.columns)
            df = df.append(a_series, ignore_index=True)  
    
    return df

def plot_gp(ch = 1, iid = "Sz2",
            path_h5 = "/mnt/archgen/users/hringbauer/data/lango.2021.may/hdf5/ch",
            figsize = (9,2.5),
            color = "maroon", 
            savefolder=""):
    """Plot Max. Genotype Posteriors for
    [iid] on [ch]."""
    ### Load the data
    with h5py.File(f"{path_h5}{ch}.h5", "r") as f: # Load for Sanity Check. See below!
        print(list(f["variants"]))
        print(np.shape(f["calldata/GT"]))
        gt = f["calldata/GT"][:]
        gp = f["calldata/GP"][:]
        samples = f["samples"][:].astype("str")
        cm = f["variants/MAP"][:]
        #iids_found = f["samples"][:].astype("str")
    #print(f"Loaded {len(samples)} Individuals.")
    idx = (samples==iid)
    assert(np.sum(idx)==1)
    i = np.where(idx)[0][0]
    y = np.max(gp[:,i,:], axis=1)

    ### Do the plotting
    plt.figure(figsize=figsize)
    ax = plt.gca()
    ax.plot(cm, y, "o", c=color, ms=1, alpha=0.5)
    ax.set_ylabel("Max. GP")
    ax.set_xlabel("Map Position [centimorgan]")
    ax.set_title(f"{iid}, Chromosome {ch}")
    ax.set_ylim([0.42, 1.02])

    if len(savefolder)>0:
            savepath = os.path.join(savefolder, f"{iid}_ch{ch}.png")
            plt.savefig(savepath, bbox_inches ='tight', pad_inches = 0, dpi=400)
            print(f"Saved to {savepath}")

    plt.show()



"""
Function to do statistics on individual IBD dataframes
In particular functions to get population level IBD rates,
normalized per pair.
@ Author: Harald Ringbauer, 2020, All rights reserved
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import itertools as it
from scipy.special import gammaincinv

def new_columns(df, df_meta, col="New Clade", col_new="", match_col="iid"):
    """Maps Entries from meta dataframe onto the IBD dataframe.
    Return modified dataframe"""
    if isinstance(col, str):
        col = [col]
     
    for c in col:
        if len(col_new)==0:
            col_new1=c
    
        dct = pd.Series(df_meta[c].values, index=df_meta[match_col]).to_dict()
        for i in range(1,3):    
            df[col_new1 + str(i)] =df["iid" + str(i)].map(dct)
            df[col_new1 + str(i)] =df["iid" + str(i)].map(dct)
    return df

#################################################################################
### Helper Stats to find relatives

def find_relatives(df, iid="", min_ibd=20, cm=20):
    """ Identify all relatives of a given sample"""
    idx = (df["iid1"]==iid) | (df["iid2"]==iid)
    df1 = df[idx]
    df1 = df1[df1[f"sum_IBD>{cm}"]>=min_ibd]
    df1 = df1.sort_values(by=f"sum_IBD>{cm}", ascending=False)
    df1 = df1.copy().reset_index(drop=True)
    return df1

def rc_date(age):
    """Ascertain whether age is rc.
    age: Can be Array"""
    rc = np.mod(age, 10)
    rc = rc>0
    return rc

def plot_age_diff(df,figsize=(8,8), title="", xlim=[-2000,2000], ylim=[10,5000], 
                  cs=["red", "green"], fs=14, yscale="log", rcdate=False):
    """Plot the Age Difference between two samples"""
    df["delta_age"] = df["age1"] - df["age2"]
    if not rcdate:
        rc = rc_date(df["age1"]) &  rc_date(df["age2"])
    else:
        rc = (df["RC_age_min1"]!=0) & (df["RC_age_min2"]!=0)
    
    colors = [cs[1] if r else cs[0] for r in rc]
    
    plt.figure(figsize=figsize)
    ax = plt.gca()
    ax.scatter(df["delta_age"], df["sum_IBD>20"], ec="k", lw=0.5, 
               color=colors)

    ax.set_yscale(yscale)
    ax.set_ylabel("Sum ROH>20 cM [cM]", fontsize=fs)
    ax.set_xlabel("Age Difference of pair [years]", fontsize=fs)
    
    #
    p1 = mpatches.Patch(color=cs[0], label="Missing RC Date")
    p2 = mpatches.Patch(color=cs[1], label='Both RC Date')
    ax.legend(handles=[p1, p2], fontsize=fs)
    
    if len(title)>0:
        ax.set_title(title)
    if len(xlim)>0:
        ax.set_xlim(xlim)
    if len(ylim)>0:
        ax.set_ylim(ylim)
    plt.show()

    
#################################################################################

def give_sub_df(df, pop1="La Caleta", pop2="La Caleta", col="clst", 
                output=True, exact=False):
    """Return sub dataframe where pair across pop1 and pop2"""
    if exact:
        idx1 = (df[col+"1"]==pop1) & (df[col + "2"]==pop2)
        idx2 = (df[col+"1"]==pop2) & (df[col + "2"]==pop1)
    else:
        idx1 = (df[col+"1"].str.contains(pop1) & df[col + "2"].str.contains(pop2))
        idx2 = (df[col+"1"].str.contains(pop2) & df[col + "2"].str.contains(pop1))
        
    idx = idx1 | idx2
    if output:
        print(f"Found: {np.sum(idx)} Pairs fitting in dataframe.\n")
    return df[idx]

def remove_iids(df, iids=[]):
    """Remove sample from pw. Indivdiual IID dataframe if pair contains one of flagged samples.
    Return Updated Dataframe.
    df: Input Dataframe
    iids: List of IIDs to remove"""
    assert(("iid1" in df.columns) and ("iid2" in df.columns))
    idx = (df["iid1"].isin(iids)) | (df["iid2"].isin(iids))
    df = df[~idx].reset_index(drop=True)
    print(f"Dropped {np.sum(idx)}/{len(idx)} pairs")
    return df

def give_stats_cm_bin(df, cms = [8,12,16,20], binary=True, output=True):
    """Return counts of IBD in bins. 
    df: IBD dataframe df of pw. individuals
    cms: Which bins to look into
    binary: Only count existing (e.g. 1/0 values)
    If upper bound 0: Take infinite bin"""
    counts = np.zeros(len(cms)-1) # Where the counts will be stored
    n = len(df) # The length of df (pw. comparisons)
    
    for i in range(len(cms)-1):
        cm1, cm2 = cms[i], cms[i+1]
        if cm2==0: # if no upper bound
            col1 = f"n_IBD>{cm1}"
            cts = df[col1].values
        else:
            col1, col2 = f"n_IBD>{cm1}", f"n_IBD>{cm2}"
            cts = df[col1].values - df[col2].values
        if binary:
            cts = np.clip(cts,0,1) # Only count existence
        tot_cts = np.sum(cts)
        counts[i] = tot_cts
        frac = np.mean(cts)
        if output:
            print(f"#IBD in {cm1}-{cm2} cM: {tot_cts}/{n}: {frac:.2f}")
    return counts

def get_IBD_stats(df, pop1="", pop2="", col="clade", exact=False,
                  cms=[4,6,8,10,12], binary=True, output=False, a=0.05):
    """Get IBD fraction statistics.
    a: Signficance level
    binary: Only count existence of IBD, not total count [0/1 per pair]
    Return fractions, confidence intervalls as well as 
    number of pairsise comparisons"""
    if len(pop1)>0 or len(pop2)>0:
        df1 = give_sub_df(df, pop1=pop1, pop2=pop2, col=col, 
                          output=output, exact=exact)
    else:
        df1 = df
    
    if len(df1)==0: # If not enough pairs for comparison
        k = len(cms)-1
        fracs, cis, n = [0]*k, np.zeros((k,2)), 0 # Empty Comparison
        if len(df1)==0 and output:
            print(f"Empty Comparison: {pop1}-{pop2}")
    else:
        counts = give_stats_cm_bin(df1, output=output, cms=cms, binary=binary)
        n = len(df1) # The number of pw. comparisons
        fracs = counts/n
        cis = get_ci_counts(counts, n, a=a)
    #stds = np.sqrt(counts)/n
    return fracs, cis, n

def get_IBD_stats_pops(df, pops1=[], pops2=[], col="clade", 
                       cms=[4,6,8,10,12], output=False, 
                       binary=True, exact=False, a=0.05):
    """Get IBD fraction statistics for list of pop pairs.
    Return lists of fractions, confidence intervalls as well as 
    number of pairwise comparisons.
    a: Significance Level
    binary: Only count existence [0/1 possible per pair]"""
    assert(len(pops1)==len(pops2)) # Sanity Check
    fracss, ciss, ns = [], [], []
    
    for i in range(len(pops1)):
        fracs, cis, n = get_IBD_stats(df, pop1=pops1[i], pop2=pops2[i], col=col,
                                      binary=binary, exact=exact, cms=cms, output=output, a=a)
        fracss.append(fracs)
        ciss.append(cis)
        ns.append(n)
    return fracss, ciss, ns

def get_ci_counts(counts, n, a=0.05, minc=1e-4):
    """Get Confidence Intervalls from counts and
    trials. Return list of CIS (lentght 2 each)
    counts: Array of Counts
    n: Total number of trials
    a: Signficance level"""
    cis = []
    for k in counts:
        if k==0:  # In case of no Count
            c0=minc/n
        else:
            c0 = gammaincinv(k, 0.5 * a)     # Lowe Bound
        c1 = gammaincinv(k + 1, 1 - 0.5 * a) # Upper bound
        cis.append([c0/n,c1/n])
    cis = np.array(cis) # Transform to numpy
    return cis

### Calulate pairwise IBD for all pops
def ibd_stats_pop_pairs(df, pop_pairs, cms = [8, 12, 16, 20, 0],
                        col = "label_region", a=0.32, binary=True,
                        output=False, exact=True):
    """For a list of population pairs, get ibd summary statistics for all pairs of
    individuals.
    Return list of ractions, list of CIs, list of #comparisons"""
    fracs, cis, ns = [], [], []
    
    for p1,p2 in pop_pairs:    
        frac, ci, n  = get_IBD_stats(df, pop1=p1, pop2=p2, col=col, binary=binary,
                                         cms=cms, output=output, a=a, exact=exact)
        fracs.append(frac)
        cis.append(ci)
        ns.append(n)
    return fracs, cis, ns

def create_ibd_pop_pair_df(pop_pairs, ns, fracs, cis, 
                           cms=[8, 12, 16, 20, 0], savepath="", ):
    """Save Population Pair dataframe if needed"""
    df_ibd = pd.DataFrame({"s1": pop_pairs[:,0], "s2": pop_pairs[:,1]})
    df_ibd["n_pairs"] = ns

    for i in range(len(cms)-1):
        cm1, cm2 = cms[i], cms[i+1]
        df_ibd[f"{cm1}-{cm2}cm"] = fracs[:,i]
        df_ibd[f"{cm1}-{cm2}cm_ci1"] = cis[:, i, 0]
        df_ibd[f"{cm1}-{cm2}cm_ci2"] = cis[:, i, 1]

    ### Save the F_ST-Matrix as well as the populaition:
    if len(savepath)>0:
        df_ibd.to_csv(savepath, sep="\t", index=False) # Save the Values
        print(f"Saved {len(df_ibd)} pw. pop comparisons.")
    return df_ibd


import numpy as np
import pandas as pd
import os as os
import sys as sys
import h5py


def get_af(f, min_gp=0.99, af_def=0.5):
    """Get Allele Frequency from HDF file f (using GP)
    min_gp: Minimum Genotype Probability used for calculation.
    af_def: Default Allele Frequency if no data."""
    gp_max = np.max(f["calldata/GP"], axis=2)
    gp_good = (gp_max>=min_gp) # The decent genotype probabilitiies
    gp_max = 0 # Delete GP max (unnecessary now)
    
    gt1 = np.sum(f["calldata/GT"], axis=2)/2.0 # Get the genotype sum
    gp_good_c = np.sum(gp_good, axis=1)
    af = np.sum(gt1 * gp_good, axis=1) #/ gp_good_c
    
    idx = gp_good_c>0
    af[idx] = af[idx] / gp_good_c[idx]
    af[~idx] = af_def # Default Value
    return af

def get_af1000G(f):
    """Get Allele Frequency - ASSUME ALL GT are set!"""
    gt1 = np.sum(f["calldata/GT"], axis=2)/2.0 # Get the genotype sum
    af = np.mean(gt1, axis=1)
    return af

def merge_in_af(path_h5, af, col_af="AF_ALL"):
    """Merge in AF into hdf5 file. Save modified h5 in place 
    af: Array of allele frequencies to save"""
    
    ### Now create the new column in hdf5
    print("Adding Allele Frequencies to HDF5...")
    with h5py.File(path_h5, 'a') as f0:
        group = f0["variants"]
        l = len(f0["variants/POS"]) # Get number of markers
        print(f"Loaded {len(af)} variants.")
        assert(l==len(af)) # Sanity Checks
        assert(np.min(af)>=0)
        assert(np.max(af)<=1)
        
        group.create_dataset(col_af, (l,), dtype='f')   
        f0[f"variants/{col_af}"][:] = af[:]
    print(f"Finshed merged in allele frequencies into {path_h5}")
    
#################################################################
### Bring over AF

def lift_af(h5_target, h5_original, field="variants/AF_ALL", 
            match_col="variants/POS", dt=np.float64, p_def=0.5):
    """Bring over field from one h5 to another. Assume field does not exist in target
    h5_original: The original hdf5 path
    h5_target: The target hdf5 path
    field: Which fielw to copy over 
    p_def: Default Value of allele frequency
    """
    g= h5py.File(h5_original, "r") # To read ref data
    f = h5py.File(h5_target, 'a') # To append data
    
    l = len(f[match_col]) #  nr target loci
    p = p_def * np.ones(l, dtype=dt) # Default value for p
    
    ### Match on position and lift over.
    its, i1, i2 = np.intersect1d(g[match_col][:], f[match_col][:], return_indices=True)
    p[i2] = g[field][:][i1]
    print(f"Intersection {len(i2)} out of {l} target HDF5 SNPs")
    
    ### Crate and write the field in target h5
    if not field in f:
        f.create_dataset(field, (l,), dtype=dt)  
    else:
        print(f"Found pre-existing field.")
    f[field][:] = p
    
    f.close()   ### Close hdf5 files
    g.close()
    
def lift_af_df(h5_target, path_df, field="variants/AF_ALL",
               match_col="variants/POS", dt=np.float64, p_def=0.5):
    """Load allele frequencies from dataframe at path_df [string]
    and merge into hdf5 file at h5_target [string] at field [string]
    Match positions on match_col [string]"""
    
    df = pd.read_csv(path_df, sep="\t") # Load Dataframe
    
    ### Load Target HDF5 
    f = h5py.File(h5_target, 'a') # Load target hdf5
    l = len(f[match_col]) #  nr target loci
    p = p_def * np.ones(l, dtype=dt) # Default value for p
    
    its, i1, i2 = np.intersect1d(df["pos"][:], f[match_col][:], return_indices=True)
    p[i2] = df["af"][:][i1]
    print(f"Intersection {len(i2)} out of {l} target HDF5 SNPs. {l-len(i2)} SNPs set to AF={p_def}")
    
    ### Crate and write the field in target h5
    if not field in f:
        f.create_dataset(field, (l,), dtype=dt)  
    else:
        print(f"Found pre-existing field.")
    f[field][:] = p
    
    f.close()  
    
    
#################################################################

def merge_in_ld_map(path_h5, path_snp1240k, chs=range(1,23), write_mode="a"):
    """Merge in MAP from eigenstrat .snp file into
    hdf5 file. Save modified h5 in place 
    path_h5: Path to hdf5 file to modify.
    path_snp1240k: Path to Eigenstrat .snp file whose map to use
    chs: Which Chromosomes to merge in HDF5 [list].
    write_mode: Which mode to use on hdf5. a: New field. r+: Change Field"""
    with h5py.File(path_h5, "r") as f:
        print("Lifting LD Map from eigenstrat to HDF5...")
        print("Loaded %i variants." % np.shape(f["calldata/GT"])[0])
        print("Loaded %i individuals." % np.shape(f["calldata/GT"])[1])

        ### Load Eigenstrat
        df_snp = pd.read_csv(path_snp1240k, header=None, sep=r"\s+", engine="python")
        df_snp.columns = ["SNP", "chr", "map", "pos", "ref", "alt"]

        rec = np.zeros(len(f["variants/POS"]))  # Create the array for vector

        for ch in chs:
            df_t = df_snp[df_snp["chr"] == ch]
            print(f"Loaded {len(df_t)} Chr.{ch} 1240K SNPs.")

            idx_f = f["variants/CHROM"][:].astype("str")==str(ch)
            if np.sum(idx_f)==0:  # If no markers found jump to next chromosome
                print("Did not find any markers...")
                continue
            rec_ch = np.zeros(len(idx_f), dtype="float")

            ### Intersect SNP positions
            its, i1, i2 = np.intersect1d(f["variants/POS"][idx_f], df_t["pos"], return_indices=True)

            l = np.sum(idx_f)
            print(f"Intersection {len(i2)} out of {l} HDF5 SNPs")

            ### Extract Map positions
            rec_ch[i1] = df_t["map"].values[i2]  # Fill in the values in Recombination map

            ### Interpolate if Needed (map position still 0)
            itp_idx = (rec_ch == 0)
            if np.sum(itp_idx) > 0:   # In case we have to interpolate
                print(f"Interpolating {np.sum(itp_idx)} variants.")
                x = df_t["pos"] 
                y = df_t["map"]   
                x1 = f["variants/POS"][:][idx_f]  # Extract all positions of interest
                rec_ch = np.interp(x1, x, y) 
            
            ### Make sure that sorted
            assert(np.all(np.diff(rec_ch)>=0))  # Assert the Recombination Map is sorted! (no 0 left and no funky stuff)
            rec[idx_f]=rec_ch # Set the Map position for chromosome indices
            print(f"Finished Chromosome {ch}.")
    
    ### Now create the new column in hdf5
    print("Adding map to HDF5...")
    with h5py.File(path_h5, write_mode) as f0:
        group = f0["variants"]
        l = len(f0["variants/POS"])
        if write_mode == "a":  # If appending new data
            group.create_dataset('MAP', (l,), dtype='f')   
        f0["variants/MAP"][:] = rec[:]

#################################################################
### Save HDF5


def save_h5(gt, ad, ref, alt, pos, 
            rec, samples, path, gp=[],
            compression="gzip", ad_group=True, gt_type="int8"):
    """Create a new HDF5 File with Input Data.
    gt: Genotype data [l,k,2]
    ad: Allele depth [l,k,2]
    ref: Reference Allele [l]
    alt: Alternate Allele [l]
    pos: Position  [l]
    m: Map position [l]
    samples: Sample IDs [k].
    Save genotype data as int8, readcount data as int16.
    ad: whether to save allele depth
    gt_type: What genotype data type save"""

    l, k, _ = np.shape(gt)  # Nr loci and Nr of Individuals

    if os.path.exists(path):  # Do a Deletion of existing File there
        os.remove(path)

    dt = h5py.special_dtype(vlen=str)  # To have no problem with saving
    with h5py.File(path, 'w') as f0:
        ### Create all the Groups
        f_map = f0.create_dataset("variants/MAP", (l,), dtype='f')
        if ad_group:
            f_ad = f0.create_dataset("calldata/AD", (l, k, 2), dtype='int8', compression=compression)
        f_ref = f0.create_dataset("variants/REF", (l,), dtype=dt)
        f_alt = f0.create_dataset("variants/ALT", (l,), dtype=dt)
        f_pos = f0.create_dataset("variants/POS", (l,), dtype='int32')
        f_gt = f0.create_dataset("calldata/GT", (l, k, 2), dtype=gt_type, compression=compression)
        if len(gp)>0:
            f_gp = f0.create_dataset("calldata/GP", (l, k, 3), dtype="f", compression=compression)     
        f_samples = f0.create_dataset("samples", (k,), dtype=dt)

        ### Save the Data
        f_map[:] = rec
        if ad_group:
            f_ad[:] = ad
        f_ref[:] = ref.astype("S1")
        f_alt[:] = alt.astype("S1")
        f_pos[:] = pos
        f_gt[:] = gt
        if len(gp)>0:
            f_gp[:] = gp
        f_samples[:] = np.array(samples).astype("S10")

    print(f"Successfully saved {k} individuals to: {path}")


"""
Class for post-processing pw. IBD list into a single 
summary dataframe for each pair of individuals, for various 
IBD length classes
@ Author: Harald Ringbauer, 2020
"""

import numpy as np
import pandas as pd
import os as os

#########################################################

def filter_ibd_df(df, min_cm=4, snp_cm=60, output=True):
    """Post Process ROH Dataframe. Filter to rows that are okay.
    min_cm: Minimum Length in CentiMorgan
    snp_cm: How many SNPs per CentiMorgan"""

    # Filter for Length:
    length_okay = (df["lengthM"] * 100) > min_cm
    
    # Filter for SNP Density:
    densities = df["length"] / (df["lengthM"] * 100)
    densities_ok = (densities > snp_cm)
    df["SNP_Dens"] = densities
    
    df = df[densities_ok & length_okay].copy()
    
    if output==True:
        print(f"> {min_cm} cM: {np.sum(length_okay)}/{len(length_okay)}")
        print(f"Of these with suff. SNPs per cM> {snp_cm}: \
              {np.sum(densities_ok & length_okay)}/{np.sum(length_okay)}")   
    return df

def roh_statistic_df(df, min_cm=0, col_lengthM="lengthM"):
    """Gives out Summary statistic of ROH df"""
    if len(df)==0:   # If no ROH Block found
        max_roh, sum_roh, n_roh = 0, 0, 0
    
    else:
        idx = df["lengthM"]>min_cm/100.0 # convert to Morgan
        if np.sum(idx)==0:
            max_roh, sum_roh, n_roh = 0, 0, 0
            
        else:
            l = df["lengthM"][idx]
            max_roh = np.max(l)
            sum_roh = np.sum(l)
            n_roh = len(l) 
    return sum_roh, n_roh, max_roh

def roh_statistics_df(df, min_cms=[8, 12, 16, 20], col_lengthM="lengthM"):
    """Gives out IBD df row summary statistics.
    Return list of sum_roh, n_roh, max_roh for each of them [as list]
    min_cm: List of minimum IBD lengths [in cM]"""  
    res = [roh_statistic_df(df, c, col_lengthM) for c in min_cms]
    return res

def create_ind_ibd_df(ibd_data = "/n/groups/reich/hringbauer/git/yamnaya/output/ibd/v43/ch_all.tsv",
                      min_cms = [8, 12, 16, 20], snp_cm = 220, 
                      min_cm = 6, sort_col = -1,
                      savepath="", output=True):
    """Create dataframe with summary statistics for each individual.
    Return this novel dataframe in hapROH format [IBD in cM]
    ibd_data: If string, what ibd file to load. Or IBD dataframe.
    savepath: If given: Save post-processed IBD dataframe to there.
    min_cms: What IBD lengths to use as cutoff in analysis [cM].
    snp_cm: Minimum Density of SNP per cM of IBD block.
    sort_col: Which min_cms col to use for sort. If <0 no sort conducted."""
    if isinstance(ibd_data, str): 
        df_ibd = pd.read_csv(ibd_data, sep="\t")
    else:
        df_ibd = ibd_data
        
    ### Filter. Be aggressive here for good set for  relatedness. Cutoff the slack
    df_ibd = filter_ibd_df(df_ibd, min_cm=min_cm, snp_cm=snp_cm, output=output) 
    
    ### Fish out the no IBD data
    if len(df_ibd)==0:
        df_res = pd.DataFrame(columns=['iid1','iid2', "max_IBD"])
        for m_cm in min_cms:
            df_res[f"sum_IBD>{m_cm}"] = 0
            df_res[f"n_IBD>{m_cm}"] = 0
    
    else: # In case there are IBD
        if output:
            print(df_ibd["ch"].value_counts())

        #df_ibd = df_ibd.sort_values(by = ["iid1", "iid2", "ch"]) # Sort so it is easier to split up
        grouped = df_ibd.groupby(['iid1','iid2'])

        df_res = pd.DataFrame(grouped.groups.keys())

        df_res.columns = ["iid1", "iid2"]

        ### Create the actual statistics
        stats = [roh_statistics_df(df, min_cms=min_cms) for _, df in grouped]
        stats = np.array(stats)

        df_res["max_IBD"] = stats[:, 0, 2] * 100

        ### Add Values for varying cM cutoffs:
        for j, m_cm in enumerate(min_cms):
            df_res[f"sum_IBD>{m_cm}"] = stats[:,j, 0] * 100
            df_res[f"n_IBD>{m_cm}"] = stats[:,j,1]

        if sort_col>=0:
            df_res = df_res.sort_values(by=f"sum_IBD>{min_cms[sort_col]}", ascending=False)  # Sort output by minimal Cutoff  
    
    ### Save if needed
    if len(savepath) > 0:
        df_res.to_csv(savepath, sep="\t", index=False)
        print(f"Saved {len(df_res)} individual IBD pairs to: {savepath}")

    return df_res

def ind_all_ibd_df(path_ibd = "/n/groups/reich/hringbauer/git/yamnaya/output/ibd/v43/ch_all.tsv",
                   col_lengthM="lengthM", snp_cm = 220, min_cm = 5,
                   output=True, sort=True, decimals=2, col_new="ibd", savepath=""):
    """Create dataframe with all IBD for each indivdiual pair
    Return this novel dataframe in hapROH format [IBD in cM]
    path_ibd: What ibd file to load.
    snp_cm: Minimum Density of SNP per cM of IBD block.
    sort: If True sort by longest IBD
    decimals: To how many decimals to round"""
    df_ibd = pd.read_csv(path_ibd, sep="\t")
    df_ibd = filter_ibd_df(df_ibd, min_cm=min_cm, snp_cm=snp_cm, output=output) # Cut the Slack
    
    if output:
        print(df_ibd["ch"].value_counts())

    grouped = df_ibd.groupby(['iid1','iid2']) # Group by Length of IBD

    df_res = pd.DataFrame(grouped.groups.keys())
    df_res.columns = ["iid1", "iid2"]

    ### Create the actual statistics
    stats = [ibd_lengths(df, col_lengthM=col_lengthM, string=True, sort=True, decimals=decimals) for _, df in grouped]
    df_res[col_new] = stats
    
    ### Sort by max IBD
    if sort:
        ibd_max = df_res[col_new].str.split(",").str[0].astype("float") # Get the longest IBD
        idx = np.argsort(-ibd_max) # Return sorted from highest to lowest
        df_res = df_res.iloc[idx,:]  # Sort output by minimal Cutoff  
        df_res=df_res.copy().reset_index(drop=True) # Reset Index
    
    ### Save if savepath given
    if len(savepath) > 0:
        df_res.to_csv(savepath, sep="\t", index=False)
        print(f"Saved {len(df_res)} individual IBD pairs to: {savepath}")

    return df_res

def ibd_lengths(df, col_lengthM="lengthM", string=True, sort=True, decimals=2, mpl=100):
    """Returns list of IBD lengths in IBD dataframe df. [in cM]
    string: If True - return comma seperated string.
    sort: Whether to sort IBD list"""  
    res = df[col_lengthM].values * mpl
    if sort:
        res = -np.sort(-res) # top to bottom
    if decimals>-np.inf:
        res = np.around(res, decimals=decimals)
    if string:
        res = ",".join(res.astype("str"))
    return res

def all_pairs_ibd(df_res, df_iid):
    """Return a new IBD dataframe with all possible IID pairs,
    set to 0 IBD if not in IBD dataframe
    df_iid: where to take iids from
    df_res: IBD dataframe (standard format)"""

    pairs = np.array(list(it.combinations(df_iid["iid"],2)))
    df_all = pd.DataFrame({'iid1':pairs[:,0], 'iid2':pairs[:,1]})
    df_all[df_res.columns[2:]]=0  ## Set new entries to 0
    df_all.set_index(["iid1", "iid2"], inplace=True)
    df_all.update(df_res.set_index(["iid1", "iid2"]))
    df_all.update(df_res.set_index(["iid2", "iid1"]))
    df_all.reset_index()
    return df_all


def combine_all_chroms(chs=[], folder_base="PATH/ch", 
                       path_save = "PATH/ch_all.tsv"):
    """Combine All Chromosomes.
    chs: Which Chromosomes to run [list]
    folder_base: Where to load from (path part up to including ch)
    path_save: Where to save the combined file to."""
    if len(chs)==0:
        chs = range(1,23)
                       
    df_ibds = []
    for ch in chs:
        path_load = f"{folder_base}{ch}.tsv"
        df = pd.read_csv(path_load, sep="\t")
        print(f"Chromosome {ch}; Loaded {len(df)} IBD")
        df_ibds.append(df)
    df_ibds = pd.concat(df_ibds)

    ### Save IBD
    df_ibds.to_csv(path_save, sep="\t", index=False) # Save the Values
    print(f"Saved {len(df_ibds)} IBD to {path_save}.")




import matplotlib.pyplot as plt
import numpy as np
import h5py as h5py
import pandas as pd

from matplotlib import rcParams
rcParams['font.family'] = 'sans-serif'   # Set the default
rcParams['font.sans-serif'] = ['Arial']  # Make sure to have the font installed (it is on cluster for Harald)



def plot_scatter_ibd(df_ibds, bins=[], min_cm=12, 
                     xlim=[-5,3600], ylim=[0,50], 
                     title="", savepath=""):
    """Plot Scatter Plot of IBD"""

    if len(bins)==0:
        bins = np.linspace(0,3600,73)

    plt.figure(figsize=(8,8))
    ax = plt.gca()
    ax.scatter(df_ibds[f"sum_IBD>{min_cm}"], df_ibds[f"n_IBD>{min_cm}"], s=40,
               ec="k", linewidth=0.5, color="deepskyblue")

    ax.set_xlabel(f"Sum IBD >{min_cm}cM [cM]", fontsize=14)
    ax.set_ylabel(f"n IBD >{min_cm}cM [cM]", fontsize=14)
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    
    if len(savepath)>0:
        plt.savefig(savepath, bbox_inches ='tight', pad_inches = 0, dpi=400)
        print(f"Saved to {savepath}")
    if len(title)>0:
        ax.set_title(title)
    plt.show()
    
    
def plot_sactter_ibd_target(df_ibds, df_target=[], bins=[], min_cm=12, s=40,
                            xlim=[-5,3600], ylim=[0,50], c_target = "maroon", c_back="gray",
                            title="", savepath=""):
    """Plot Scatter plot of sum and number of IBD segments (x & y axis).
    df_ibds: Background plot
    df_target: Foreground plot"""
    
    if len(bins)==0:
        bins = np.linspace(0,3600,73)

    plt.figure(figsize=(8,8))
    ax = plt.gca()
    ax.scatter(df_ibds[f"sum_IBD>{min_cm}"], df_ibds[f"n_IBD>{min_cm}"], s=s,
               ec="k", linewidth=0.3, color=c_back, zorder=0)
    
    ### Plot Foreground
    ax.scatter(df_target[f"sum_IBD>{min_cm}"], df_target[f"n_IBD>{min_cm}"], s=s*2,
               ec="k", linewidth=0.8, color=c_target, zorder=10)

    ax.set_xlabel(f"Sum IBD >{min_cm}cM [cM]", fontsize=14)
    ax.set_ylabel(f"n IBD >{min_cm}cM [cM]", fontsize=14)
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    
    if len(savepath)>0:
        plt.savefig(savepath, bbox_inches ='tight', pad_inches = 0, dpi=400)
        print(f"Saved to {savepath}")
    if len(title)>0:
        ax.set_title(title)
    plt.show()
    


"""ancIBD: Main HMM class.
Pulls together other ancIBD modules and directs loading/running them.
@author: Harald Ringbauer, 2020
"""


import os as os
import sys as sys   # For piping the output
import numpy as np
from ancIBD.emission import load_emission_model
from ancIBD.transition import load_transition_model
from ancIBD.loaddata import load_loaddata
from ancIBD.hmm import load_fwd_bwd_func
from ancIBD.postprocessing import load_Postprocessing
from time import time # All I need is... time

class HMM_Full(object):
    """Analyze Class for HMMs. Wrapper for various subclasses, making them
    work together.
    Contains objects as field for pre-processing, transition as well as
    emission probabilities, loading and post-processing
    Contains most Parameters (but some of them like the output folders are decided
    by pre-processing subclass)"""
    folder_in = ""
    folder_out = "" # For book keeping of the output path
    
    l_model = "" # simulated
    e_model = "" # haploid_gl
    t_model = "" # standard
    p_model = ""
    h_model = "" # FiveState FiveStateFast FiveStateScaled
    post_model = ""
    
    l_obj, t_obj, e_obj = 0, 0, 0 # Placeholder for the objects
    fwd_bwd = 0 # Placeholder Forward Backward Function
    submat33 = True
    in_val = 1e-4 # The prior forward prob. in first IBD state
    output = True
    

    def __init__(self, folder_in="", l_model="simulated", t_model="standard", 
                 e_model="haploid_gl", h_model = "FiveStateFast", p_model="hapROH",
                 output=True, load=True):
        """Initialize Class.
        Keywords for which subclasses to load (via factory functions from there.
        load: Whether to load all models at intialization. True per Default."""
        self.set_params(folder_in=folder_in, l_model=l_model, 
                        t_model=t_model, e_model=e_model, 
                        h_model=h_model, p_model=p_model, output=output)
        if load:
            self.load_objects()

    def load_objects(self):
        """Load all the required Objects in right order"""
        self.l_obj = load_loaddata(l_model = self.l_model, path=self.folder_in)
        self.t_obj = load_transition_model(t_model = self.t_model)
        self.e_obj = load_emission_model(e_model = self.e_model)
        self.p_obj = load_Postprocessing(p_model=self.p_model)
        self.fwd_bwd = load_fwd_bwd_func(h_model=self.h_model)
        
        ### pass on output parameter
        for o in [self.l_obj, self.t_obj, self.e_obj, self.p_obj]:
            o.set_params(output=self.output)
    
    
    def run_fwd_bwd(self, full=True):
        """Run Forward Backward algorithm."""
        htsl, p, r_vec, _ =  self.l_obj.load_all_data()
        e_mat = self.e_obj.give_emission_matrix(htsl, p)
        t_mat = self.t_obj.full_transition_matrix(r_vec, n=4, submat33=self.submat33)
        
        if full:
            post, fwd, bwd, tot_ll = self.fwd_bwd(e_mat, t_mat, in_val = self.in_val, 
                                                  full=full, output=self.output)
            self.p_obj.call_roh(r_vec, post)
            return post, r_vec, fwd, bwd, tot_ll
        else:
            post = self.fwd_bwd(e_mat, t_mat, in_val = self.in_val, 
                                full=full, output=self.output)
            self.p_obj.call_roh(r_vec, post)
            return post, r_vec  
    
    ##################################################
    ### Run FWD-BWD timed [LEGACY code]
    
    def run_fwd_bwd_timed(self, full=True):
        """Run Forward Backward algorithm.
        Legacy: Timed version."""
        t = time()
        htsl, p, r_vec, _ =  self.l_obj.load_all_data()
        e = time()
        print(f"Runtime Loading: {(e-t)} s")
        
        t = time()
        e_mat = self.e_obj.give_emission_matrix(htsl, p)
        e = time()
        print(f"Runtime E Mat.: {(e-t)} s")
        
        t = time()
        t_mat = self.t_obj.full_transition_matrix(r_vec, n=4, submat33=self.submat33)
        e = time()
        print(f"Runtime T Mat.: {(e-t)} s")
        
        if full:
            post, fwd, bwd, tot_ll = self.fwd_bwd(e_mat, t_mat, in_val = self.in_val, 
                                                  full=full, output=self.output)
            self.p_obj.call_roh(r_vec, post)
            return post, r_vec, fwd, bwd, tot_ll
        else:
            t = time()
            post = self.fwd_bwd(e_mat, t_mat, in_val = self.in_val, 
                                full=full, output=self.output)
            e = time()
            print(f"Runtime HMM calc.: {(e-t)} s")
            t = time()
            self.p_obj.call_roh(r_vec, post)
            e = time()
            print(f"Runtime Postprocessing: {(e-t)} s")
            return post, r_vec                                                                    
                 
    def set_params(self, **kwargs):
        """Set the Parameters.
        Takes keyworded arguments"""
        for key, value in kwargs.items():
            setattr(self, key, value)
            
    ##################################################
    ### Prepping output folder and piping output there
    
    def prepare_path(self, base_path, ch, prefix_out="", logfile=False):
        """Prepare the output path and pipe printing for one Individual.
        Create Path if not already existing.
        prefix_out: Optional additonal folder.
        logfile: Whether to pipe output to log-file [WARNING: This is an ugly hack]"""  
        path_out = os.path.join(base_path, prefix_out, "")
        if not os.path.exists(path_out):
                os.makedirs(path_out)
        self.folder_out = path_out
        
        ### Activate  print to LOG FILE output if given
        if logfile:
            path_log = os.path.join(path_out, "hmm_run_log.txt")
            print(f"Set Output Log to path: {path_log}")
            sys.stdout = open(path_log, 'w') 
        return path_out
    
    def reset_print(self):
        """Resets output to console."""
        sys.stdout = sys.__stdout
        if self.output:
            print("Output was reset to standard console.")


"""
Functions to prepare HDF5 file from imputed VCFs
@ Author: Harald Ringbauer, 2021
"""

import numpy as np
import pandas as pd
import os as os
import sys as sys
import h5py
import allel
from ancIBD.IO.h5_modify import merge_in_af, get_af, lift_af_df, merge_in_ld_map # get_af1000G, lift_af
#sys.path.insert(0, "/n/groups/reich/hringbauer/git/HAPSBURG/package/")  # hack to get local package first in path
#from hapsburg.PackagesSupport.h5_python.h5_functions import merge_in_ld_map
#sys.path.append("/n/groups/reich/hringbauer/git/hapBLOCK/python3/")

def save_1240kmarkers(snp1240k_path="", marker_path="", ch=0):
    """Save all 1240 Markers of .snp eigenstrat file.
    to marker_path.
    ch: Chromosome. If null filter all of them"""
    df_snp = pd.read_csv(snp1240k_path, header=None, 
                         sep=r"\s+", engine="python")
    df_snp.columns = ["SNP", "chr", "map", 
                      "pos", "ref", "alt"]
    if ch>0:
        df_snp = df_snp[df_snp["chr"] == ch]
    print(f"Loaded {len(df_snp)} Chr.{ch} SNPs.")

    df_save = df_snp[["chr", "pos"]]
    df_save.to_csv(marker_path, sep="\t", header=None, index=False)
    print(f"Saved {len(df_save)} 1240k Markers on Chr. {ch} to {marker_path}")
    
def save_1240_1000g_kmarkers(ch=3, snp_path="", marker_path=""):
    """Save all 1240 and 1000G Markers of .snp eigenstrat file.
    to marker_path. Loads Ali Path file
    snp_path: Where to find the SNPs plus their types"""
    df1=pd.read_csv(snp_path, sep="\t", header=None)
    df1.columns=["chr", "pos", "pos1", "ref", "alt", "type"]
    print(f"Loaded {len(df1)} SNPs on Chr. {ch}")
    df2 = df1[df1["type"]=="1kg+1240k"]
    print(f"Loaded {len(df2)} Chr.{ch} SNPs in both 1240K and 1000G.")
    df_save = df2[["chr", "pos1"]]
    df_save.to_csv(marker_path, sep="\t", header=None, index=False)
    print(f"Saved {len(df_save)} 1240k+1000G Markers on Chr. {ch} to {marker_path}")
    
def bctools_filter_vcf(in_vcf_path="", out_vcf_path="", marker_path=""):
    """Same as PLINK, but with bcftools and directly via Marker Positions.
    filter_iids: Whether to use the .csv with Indivdiduals to extract.
    Check whether out_vcf_path has .gz or .vcf at end and compresses for former"""
    if out_vcf_path.endswith(".gz"):
        command = f"bcftools view -Oz -o {out_vcf_path} -T {marker_path} -M2 -v snps {in_vcf_path}"
    elif out_vcf_path.endswith(".vcf"):
        command = f"bcftools view -Ov -o {out_vcf_path} -T {marker_path} -M2 -v snps {in_vcf_path}"
    else:
        raise RuntimeError("Enter valid filename for VCF: Has to end with .gz or .vcf")
    
    print(f"Running bash command: \n{command}")
    os.system(command)  # Run the actual command in shell
    
    #!bcftools view -Oz -o $out_vcf_path -T $marker_path -M2 -v snps $in_vcf_path
    print("Finished BCF tools filtering to target markers.")
    
def bctools_filter_vcf_allvariants(in_vcf_path="", out_vcf_path="", marker_path=""):
    """Same as PLINK, but with bcftools and directly via Marker Positions.
    filter_iids: Whether to use the .csv with Indivdiduals to extract"""
    command = f"bcftools view -Oz -o {out_vcf_path} -T {marker_path} -v snps {in_vcf_path}"
    os.system(command)
    print("Finished BCF tools filtering.")
    
def merge_vcfs(in_vcf_paths=[], out_vcf_path=""):
    """Merges Set of VCFs into one VCF. 
    in_vcf_paths: List of VCFs to merge
    out_vcf_path: Output of VCF"""
    paths_merge = " ".join(in_vcf_paths)
    command = f"bcftools concat -n -o {out_vcf_path} {paths_merge}"
    os.system(command)
    print("Finished BCF tools filtering.")
    
##############################################################
### Function Identical to vcf_to_hdf5.py in cluster/ folder

def vcf_to_1240K_hdf(in_vcf_path = "/n/groups/reich/ali/WholeGenomeImputation/imputed/v43.4/chr3.bcf",
                     path_vcf = "./data/vcf/1240k_v43/ch3.vcf.gz",
                     path_h5 = "./data/hdf5/1240k_v43/ch3.h5",
                     marker_path = "./data/filters/ho_snps_bcftools_ch3.csv",
                     map_path = "/n/groups/reich/DAVID/V43/V43.5/v43.5.snp",
                     af_path = "",
                     col_sample_af = "AF_SAMPLE",
                     chunk_length=10000, chunk_width=8, buffer_size=20000,
                     ch=3):
    """
    Convert Ali's vcf to 1240K hdf5. 
    
    Parameters
    ------------
    in_vcf_path: str
        Input VCF file (i.e, output from GLIMPSE)
    path_vcf: str
        A filtered vcf of in_vcf_path that contains only 1240k sites.
    path_h5: str
        Path of the output HDF5 files
    marker_path: str
        Path to file containing SNPs to downsample. If marker_path empty, no SNP filtering done.
    map_path: str 
        Path to eigenstrat SNP file containing genetic map. These are merged into a hdf5 field "variants/MAP". If map_path empty, no genetic map is merged in.
    af_path: str
        Path to tab-seperated table containing allele frequencies. There are merged into the hdf5 field "variants/AF_ALL". If no such path given, no allele frequencies are merged in.
    col_sample_af: str:
        The hdf5 column name for the allele frequency calculated from sample. If left empty, no such column will be calculated or added. 
    """ 
    print("Print downsampling to 1240K...")
    if len(marker_path)>0:
        bctools_filter_vcf(in_vcf_path = in_vcf_path,
                           out_vcf_path= path_vcf,
                           marker_path = marker_path)
    else:
        path_vcf = in_vcf_path  ### If no temporary VCF use input VCF
    
    if os.path.exists(path_h5):  # Do a Deletion of existing hdf5 File there
        print(f"Deleting previous HDF5 file at path_h5: {path_h5}...")
        os.remove(path_h5)
    
    print("Converting to HDF5...")
    allel.vcf_to_hdf5(input=path_vcf, output=path_h5, 
                      fields = ['variants/*', 'calldata/*', "samples"], 
                      types = {"samples":"S60", "calldata/GT":np.int8,
                               "calldata/GP":np.float32, "calldata/PL":np.float32}, 
                      buffer_size=buffer_size,
                      chunk_length = chunk_length, chunk_width=chunk_width,
                      compression="gzip") # Do the conversion to hdf5. Takes hours
    print("Finished conversion to hdf5!")
    
    print("Merging in LD Map..")
    if len(map_path)>0:
        merge_in_ld_map(path_h5=path_h5, 
                    path_snp1240k=map_path,
                    chs=[ch])
        
    ### Calculate and Merge in Allele Frequency
    if len(col_sample_af)>0:
        print(f"Calculating in sample allele frequencies and saving at hdf5 column {col_sample_af}")
        with h5py.File(path_h5, "r") as f:
            af = get_af(f) 
        merge_in_af(path_h5, af, col_af=col_sample_af)
        
    if len(af_path)>0:
        lift_af_df(h5_target=path_h5, path_df=af_path,
                   field='variants/AF_ALL') # The field to use for Allele Frequencies

    print(f"Transformation complete! Find new hdf5 file at: {path_h5}\n")


from setuptools import setup, find_packages
from setuptools.extension import Extension
from Cython.Distutils import build_ext

###
# We distribute the source only,
# to avoid trouble with platform

### MODIFY IF YOU WANT TO USE CYTHON FOR BUILDING:
USE_CYTHON = True   

# Code for Cython / C Extension
ext = '.pyx' if USE_CYTHON else '.c'
extensions = [Extension("ancIBD.cfunc", ["ancIBD/cfunc" + ext])]

if USE_CYTHON:
    from Cython.Build import cythonize
    extensions = cythonize(extensions)
  
with open("README.md", "r") as fh:
    long_description = fh.read()
    
setup(
    name="ancIBD",
    version="0.5",  # a means alpha
    author="Harald Ringbauer",
    author_email="harald_ringbauer@eva.mpg.de",
    description="Identify IBD segments between pairs of individuals in ancient human DNA data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hringbauer/ancIBD",
    packages=find_packages(),
    ext_modules=extensions,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['numpy', 'pandas', 'scipy', 'h5py', 
                      'psutil', 'cython', 'matplotlib', 'pysam', 'scikit-allel'],
    entry_points ={
            'console_scripts': [
                'ancIBD-summary = ancIBD.ancIBD_summary:main',
                'ancIBD-run = ancIBD.run_ancIBD:main'
            ]
        },
)
